/* -*- C++ -*-
 * $Id: flbrowser.cc,v 1.20 2001/11/26 23:42:24 ttate Exp $
 */

#include "fltk.h"
#include "flbrowser.h"

VALUE rb_cFLBrowser_;
VALUE rb_cFLBrowser;
VALUE rb_cFLHold_Browser;
VALUE rb_cFLMulti_Browser;
VALUE rb_cFLSelect_Browser;

static int no_columns[1] = {0};

static FN_S_NEW(Browser_);
WIDGET_FNS(Browser_);

static VALUE rb_FLBrowser__select(int argc, VALUE argv[], VALUE self)
{
  RBFLBrowser_ *widget = (RBFLBrowser_*)rb_to_fl(self);
  VALUE item, state, docb;
  int r;

  switch( rb_scan_args(argc, argv, "12", &item, &state, &docb) ){
  case 1:
    r = widget->select((void*)item);
    break;
  case 2:
    r = widget->select((void*)item, FIX2INT(state));
    break;
  case 3:
    r = widget->select((void*)item, FIX2INT(state), BOOL2INT(docb));
    break;
  };
  return INT2BOOL(r);
};

static VALUE rb_FLBrowser__select_only(int argc, VALUE argv[], VALUE self)
{
  RBFLBrowser_ *widget = (RBFLBrowser_*)rb_to_fl(self);
  VALUE item, docb;
  int r;

  switch( rb_scan_args(argc, argv, "11", &item, &docb) ){
  case 1:
    r = widget->select_only((void*)item);
    break;
  case 2:
    r = widget->select_only((void*)item, BOOL2INT(docb));
    break;
  };
  return INT2BOOL(r);
};


static FN_S_NEW(Browser);
WIDGET_FNS(Browser);
static FN_GET_INT(Browser,size);
static FN_BROWSER_TEXT(Browser, text);
static FN_INIT2(Browser, Fl_Group::size, {
  // replace the default value with our default value -- no_columns.
  widget->column_widths(no_columns);
  // Fltk::Browser has @items for some methods such as `add',`remove' and `clear'.
  rb_ivar_set(self, rb_id_iitems, rb_ary_new());
});
static FN_SET_INT(Browser,bottomline);
static FN_SET_INT(Browser,middleline);
static FN_GETSET_INT(Browser,column_char);
static FN_GETSET_INT(Browser,format_char);

static VALUE rb_FLBrowser_column_widths(int argc, VALUE argv[], VALUE self)
{
  RBFLBrowser *widget = (RBFLBrowser*)rb_to_fl(self);
  VALUE ary;
  int *cary;

  switch( rb_scan_args(argc,argv,"01",&ary) ){
  case 0:
    cary = (int*) widget->column_widths();
    return cary2iary(cary);
  case 1:
    Check_Type(ary, T_ARRAY); /* must not be Qnil */
    cary = (int*) widget->column_widths();
    if( cary && (cary != no_columns) ){
      free(cary);
    };
    if( RARRAY(ary)->len == 0 ){
      widget->column_widths(no_columns);
    }
    else{
      widget->column_widths(iary2cary(ary));
    };
    return Qnil;
  };
};

static VALUE rb_FLBrowser_add(int argc, VALUE argv[], VALUE self)
{
  RBFLBrowser *widget = (RBFLBrowser*)rb_to_fl(self);
  VALUE str, data;

  switch( rb_scan_args(argc, argv, "11", &str, &data) ){
  case 1:
    widget->add(STR2CSTR(str));
    rb_ary_push(rb_ivar_get(self, rb_id_iitems), Qnil);
    break;
  case 2:
    widget->add(STR2CSTR(str), (void*)data);
    rb_ary_push(rb_ivar_get(self, rb_id_iitems), data);
    break;
  };
  return Qnil;
};

static VALUE rb_FLBrowser_clear(VALUE self)
{
  RBFLBrowser *widget = (RBFLBrowser*)rb_to_fl(self);
  widget->clear();
  rb_ary_clear(rb_ivar_get(self, rb_id_iitems));
  return Qnil;
};

static VALUE rb_FLBrowser_data(int argc, VALUE argv[], VALUE self)
{
  RBFLBrowser *widget = (RBFLBrowser*)rb_to_fl(self);
  VALUE line, data;
  void *ret = 0;

  switch( rb_scan_args(argc, argv, "11", &line, &data) ){
  case 1:
    ret = widget->data(NUM2INT(line));
    break;
  case 2:
    widget->data(NUM2INT(line), (void*)data);
    rb_ary_store(rb_ivar_get(self, rb_id_iitems), NUM2INT(line) - 1, data);
    break;
  };
  return ret ? (VALUE)ret : Qnil;
};

static VALUE rb_FLBrowser_remove(VALUE self, VALUE line)
{
  RBFLBrowser *widget = (RBFLBrowser*)rb_to_fl(self);
  widget->remove(NUM2INT(line));
  rb_ary_delete_at(rb_ivar_get(self, rb_id_iitems), NUM2INT(line) - 1);
  return Qnil;
};

static FN_S_NEW(Hold_Browser);
WIDGET_FNS(Hold_Browser);
static FN_GETSET_INT(Hold_Browser,value);

static FN_S_NEW(Multi_Browser);
WIDGET_FNS(Multi_Browser);
static FN_GET_BOOL_SET_INT(Multi_Browser,selected);

static FN_S_NEW(Select_Browser);
WIDGET_FNS(Select_Browser);

extern "C" void
Init_flbrowser(){
  rb_cFLBrowser_ = rb_define_class_under(rb_mFLTK, "Browser_", rb_cFLWidget);
  DEF_S_METHOD(Browser_,new,-1);
  DEF_WIDGET_METHODS(Browser_);
  DEF_METHOD(Browser_, select, -1);
  DEF_METHOD(Browser_, select_only, -1);

  rb_cFLBrowser  = rb_define_class_under(rb_mFLTK, "Browser", rb_cFLBrowser_);
  DEF_S_METHOD(Browser,new,-1);
  DEF_WIDGET_METHODS(Browser);
  DEF_METHOD(Browser, initialize, -1);
  DEF_METHOD(Browser, clear, 0);
  DEF_METHOD(Browser, add, -1);
  DEF_METHOD(Browser, remove, 1);
  DEF_METHOD(Browser, data, -1);
  DEF_METHOD(Browser, size, 0);
  DEF_METHOD(Browser, text, -1);
  DEF_METHOD(Browser, column_widths, -1);
  DEF_METHOD(Browser, column_char, -1);
  DEF_METHOD(Browser, format_char, -1);
  DEF_METHOD(Browser, bottomline, 1);
  DEF_METHOD(Browser, middleline, 1);


  rb_cFLHold_Browser  = rb_define_class_under(rb_mFLTK, "HoldBrowser",
					      rb_cFLBrowser);
  DEF_S_METHOD(Hold_Browser,new,-1);
  DEF_WIDGET_METHODS(Hold_Browser);
  DEF_METHOD(Hold_Browser, value, -1);

  rb_cFLMulti_Browser  = rb_define_class_under(rb_mFLTK, "MultiBrowser",
					       rb_cFLBrowser);
  DEF_S_METHOD(Multi_Browser,new,-1);
  DEF_WIDGET_METHODS(Multi_Browser);
  DEF_METHOD(Multi_Browser, selected, 1);

  rb_cFLSelect_Browser  = rb_define_class_under(rb_mFLTK, "SelectBrowser",
						rb_cFLBrowser);
  DEF_S_METHOD(Select_Browser,new,-1);
  DEF_WIDGET_METHODS(Select_Browser);
};
