/* -*- C++ -*-
 * $Id: flfile.cc,v 1.2 2001/11/27 08:12:34 ttate Exp $
 */

#include "fltk.h"

#if FL_MAJOR_VERSION >=1 && FL_MINOR_VERSION >= 1

#include "flfile.h"

VALUE rb_cFLFile_Browser;
VALUE rb_cFLFile_Chooser;

static FN_S_NEW(File_Browser);
WIDGET_FNS(File_Browser);
static FN_GET_INT_SET_STR(File_Browser,load);
static FN_GETSET_INT(File_Browser,filetype);
static FN_GETSET_STR(File_Browser,filter);
static FN_GETSET_INT(File_Browser,iconsize);


static FN_CALL(File_Chooser,show);
static FN_CALL(File_Chooser,hide);

static VALUE rb_FLFile_Chooser_new(VALUE klass, RBFLFile_Chooser *chooser)
{
  VALUE obj = Data_Wrap_Struct(klass,0,0,chooser);
  rb_fltk_objmap_aset((Fl_Widget*)chooser, obj);
  return obj;
};

static VALUE rb_FLFile_Chooser_s_new(int argc, VALUE argv[], VALUE klass)
{
  RBFLFile_Chooser *chooser;
  VALUE d,p,t,title;
  VALUE obj;

  switch( rb_scan_args(argc, argv, "04", &d, &p, &t, &title) ){
  case 0:
    chooser = new RBFLFile_Chooser(NULL, NULL, 0, NULL);
    break;
  case 1:
    chooser = new RBFLFile_Chooser(STR2CPTR(d), NULL, 0, NULL);
    break;
  case 2:
    chooser = new RBFLFile_Chooser(STR2CPTR(d), STR2CPTR(p), 0, NULL);
    break;
  case 3:
    chooser = new RBFLFile_Chooser(STR2CPTR(d), STR2CPTR(p), NUM2INT(t), NULL);
    break;
  case 4:
    chooser = new RBFLFile_Chooser(STR2CPTR(d), STR2CPTR(p),
				   NUM2INT(t), STR2CPTR(title));
    break;
  };
  obj = rb_FLFile_Chooser_new(klass,chooser);
  rb_obj_call_init(obj,argc,argv);
  return obj;
};

static VALUE rb_FLFile_Chooser_initialize(int argc, VALUE argv[], VALUE self)
{
  return Qnil;
};

static VALUE rb_FLFile_Chooser_destroy(VALUE self)
{
  RBFLFile_Chooser *p = (RBFLFile_Chooser*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)p);
  return Qnil;
};

static void
rb_flfilechooser_callback_func(Fl_File_Chooser *widget, void *userdata)
{
  VALUE obj;
  VALUE callback, data;

  obj = rb_fltk_objmap_aref((Fl_Widget*)widget);
  if( obj == Qnil ) return ;

  callback = rb_ivar_get(obj, rb_id_icallback);
  if( userdata ){
    data = (VALUE)userdata;
  }
  else{
    data = Qnil;
  };

  if( callback != Qnil ){
    rb_funcall(callback, rb_id_call, 2, obj, data);
  };
};

static VALUE
rb_FLFile_Chooser_callback(int argc, VALUE argv[], VALUE self)
{
  VALUE proc, data;
  RBFLFile_Chooser *widget;
  int use_data = 0;

  if( rb_block_given_p() ){
    proc = rb_f_lambda();
    switch( rb_scan_args(argc, argv, "01", &data) ){
    case 0:
      use_data = 0;
      break;
    case 1:
      use_data = 1;
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "11", &proc, &data) ){
    case 1:
      use_data = 0;
      break;
    case 2:
      use_data = 1;
      break;
    };
  };

  rb_ivar_set(self, rb_id_icallback, proc);
  widget = (RBFLFile_Chooser*)rb_to_fl(self);
  if( use_data ){
    widget->callback(rb_flfilechooser_callback_func, (void*)data);
  }
  else{
    widget->callback(rb_flfilechooser_callback_func, (void*)Qnil);
  };

  return Qnil;
};


extern "C" void Init_flfile(void)
{
  rb_cFLFile_Browser = rb_define_class_under(rb_mFLTK, "FileBrowser",
					     rb_cFLBrowser);
  DEF_S_METHOD(File_Browser,new,-1);
  DEF_WIDGET_METHODS(File_Browser);
  DEF_METHOD(File_Browser,load,1);
  DEF_METHOD(File_Browser,filetype,-1);
  DEF_METHOD(File_Browser,filter,-1);
  DEF_METHOD(File_Browser,iconsize,-1);

#define NUMCONST(c) \
  rb_define_const(rb_cFLFile_Browser,#c,INT2NUM(RBFLFile_Browser::c))
  NUMCONST(FILES);
  NUMCONST(DIRECTORIES);
#undef NUMCONST


  rb_cFLFile_Chooser = rb_define_class_under(rb_mFLTK, "FileChooser",
					     rb_cObject);
  DEF_S_METHOD(File_Chooser,new,-1);
  DEF_METHOD(File_Chooser,initialize,-1);
  DEF_METHOD(File_Chooser,destroy,0);
  DEF_METHOD(File_Chooser,callback,-1);
  DEF_METHOD(File_Chooser,show,0);
  DEF_METHOD(File_Chooser,hide,0);
#define NUMCONST(c) \
  rb_define_const(rb_cFLFile_Chooser,#c,INT2NUM(RBFLFile_Chooser::c))
  NUMCONST(SINGLE);
  NUMCONST(MULTI);
  NUMCONST(CREATE);
  NUMCONST(DIRECTORY);
#undef NUMCONST
};

#endif /* FL_MAJOR_VERSION ... */
