/* -*- C++ -*-
 * $Id: flgroup.cc,v 1.23 2001/11/25 17:04:34 ttate Exp $
 */

#include "fltk.h"
#include "flgroup.h"

VALUE rb_cFLGroup;

static FN_S_NEW(Group);
WIDGET_FNS(Group);
FN_CALL(Group,end); /* non-static function */
static FN_GROUP_INIT(Group);
static FN_SET_WIDGET(Group,add);
static FN_CALL(Group,init_sizes);
static FN_GETSET_WIDGET(Group,resizable);
static FN_SET_REFWIDGET(Group,add_resizable);
static FN_SET_REFWIDGET(Group,remove);
static FN_GET_INT_SET_WIDGET(Group,find);

VALUE rb_FLGroup_begin(VALUE self) /* isn't static function */
{
  RBFLGroup *widget = (RBFLGroup*)rb_to_fl(self);

  widget->begin();
  if( rb_block_given_p() ){
    rb_ensure(RUBY_METHOD_FUNC(rb_yield), self,
	      RUBY_METHOD_FUNC(rb_FLGroup_end), self);
    // rb_yield(self);
    // widget->end();
  };
  return Qnil;
};

static VALUE rb_FLGroup_children(VALUE self)
{
  RBFLGroup *widget = (RBFLGroup*)rb_to_fl(self);
  int n = widget->children();
  VALUE ary = rb_ary_new();
  Fl_Widget *child;
  VALUE obj;

  for( int i = 0; i < n; i++ ){
    child = widget->child(i);
    obj   = rb_fltk_objmap_aref(child);
    rb_ary_push(ary,obj);
  };
  return ary;
};

static VALUE rb_FLGroup_array(VALUE self)
{
  rb_warn("use %s#children", rb_class2name(CLASS_OF(self)));
  return rb_FLGroup_children(self);
};

static VALUE rb_FLGroup_child(VALUE self, VALUE n)
{
  RBFLGroup *widget = (RBFLGroup*)rb_to_fl(self);
  Fl_Widget *child = widget->child(NUM2INT(n));

  rb_warn("use %s#children[%d]", rb_class2name(CLASS_OF(self)), NUM2INT(n));
  if( child ){
    return rb_fltk_objmap_aref(child);
  }
  else{
    return Qnil;
  };
};

static VALUE rb_FLGroup_s_current(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1;

  switch( rb_scan_args(argc, argv, "01", &arg1) ){
  case 0:
    return rb_fltk_objmap_aref(RBFLGroup::current());
  case 1:
    if( rb_obj_is_kind_of(arg1, rb_cFLGroup) ){
      RBFLGroup::current((Fl_Group*)(rb_to_fl(arg1)));
    }
    else{
      rb_raise(rb_eTypeError, "the argument must inherits from Fltk::Group");
    };
    return Qnil;
  };
};

static VALUE rb_FLGroup_insert(int argc, VALUE argv[], VALUE self)
{
  RBFLGroup *widget = (RBFLGroup*)rb_to_fl(self);
  VALUE arg1, arg2;

  switch( rb_scan_args(argc, argv, "11", &arg1, &arg2) ){
  case 1:
    return rb_FLGroup_add(self,arg1);
  case 2:
    if( rb_obj_is_kind_of(arg2,rb_cFLWidget) ){
      widget->insert(*(rb_to_fl(arg1)), rb_to_fl(arg2));
      return Qnil;
    }
    else{
      widget->insert(*(rb_to_fl(arg1)), NUM2INT(arg2));
      return Qnil;
    };
  };
};

extern "C" void
Init_flgroup(){
  rb_cFLGroup = rb_define_class_under(rb_mFLTK, "Group", rb_cFLWidget);
  DEF_S_METHOD(Group,new,-1);
  DEF_S_METHOD(Group,current,-1);
  DEF_WIDGET_METHODS(Group);
  DEF_METHOD(Group,initialize,-1);
  RB_DEF_METHOD(rb_cFLGroup, "groupend", rb_FLGroup_end, 0);
  RB_DEF_METHOD(rb_cFLGroup, "group_end", rb_FLGroup_end, 0);
  RB_DEF_METHOD(rb_cFLGroup, "groupbegin", rb_FLGroup_begin, 0);
  RB_DEF_METHOD(rb_cFLGroup, "group_begin", rb_FLGroup_begin, 0);
  DEF_METHOD_Q(Group,resizable,-1);
  DEF_METHOD(Group,add,1);
  DEF_METHOD(Group,add_resizable,1);
  DEF_METHOD(Group,children,0);
  DEF_METHOD(Group,array,0);
  DEF_METHOD(Group,child,1);
  DEF_METHOD(Group,init_sizes,0);
  DEF_METHOD(Group,find,1);
  DEF_METHOD(Group,remove,1);
  DEF_METHOD(Group,insert,-1);
};
