/* -*- C++ -*-
 * $Id: flimage.cc,v 1.4 2001/12/04 16:58:33 ttate Exp $
 */

#include "fltk.h"
#include "flwidget.h"
#include "flimage.h"

VALUE rb_cFLImage;
VALUE rb_cFLRGB_Image;

static void rb_FLImage_free(Fl_Image *image)
{
};

static VALUE rb_FLImage_new(VALUE klass, Fl_Image *image)
{
  VALUE obj;
  obj = Data_Wrap_Struct(klass, 0, rb_FLImage_free, image);
  rb_fltk_objmap_aset((Fl_Widget*)image, obj);
  return obj;
};

VALUE rb_FLImage_initialize(int argc, VALUE argv[], VALUE self)
{
  return Qnil;
};

#if !defined(USE_RGB_IMAGE)
VALUE rb_FLImage_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, xbm, str, w, h, d, ld;
  RBFLImage *image;
  int ary_len, str_len;
  uchar *data;
  uchar *rb_fltk_create_image(VALUE);

  rb_scan_args(argc, argv, "32", &xbm,&w,&h, &d, &ld);

  Check_Type(xbm, T_ARRAY);
  data = rb_fltk_create_image(xbm);
  switch( argc ){
  case 3:
    image = new RBFLImage(data,NUM2INT(w),NUM2INT(h));
    break;
  case 4:
    image = new RBFLImage(data,NUM2INT(w),NUM2INT(h),NUM2INT(d));
    break;
  case 5:
    image = new RBFLImage(data,NUM2INT(w),NUM2INT(h),NUM2INT(d), NUM2INT(ld));
    break;
  }
  obj = rb_FLImage_new(klass, image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

VALUE rb_FLImage_destroy(VALUE self)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)image);
  if( image && image->array ){
    free((char*)(image->array));
    delete image;
  };

  return Qnil;
};

#else /* if defined(USE_RGB_IMAGE) */

VALUE rb_FLImage_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, str, w, h, d;
  RBFLImage *image;
  int ary_len, str_len;
  uchar *data;
  uchar *rb_fltk_create_image(VALUE);

  rb_scan_args(argc, argv, "30", &w,&h,&d);

  image = new RBFLImage(NUM2INT(w),NUM2INT(h),NUM2INT(d));
  obj = rb_FLImage_new(klass,image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

VALUE rb_FLRGB_Image_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, xbm, str, w, h, d, ld;
  RBFLRGB_Image *image;
  int ary_len, str_len;
  uchar *data;
  uchar *rb_fltk_create_image(VALUE);

  rb_scan_args(argc, argv, "32", &xbm,&w,&h, &d, &ld);

  Check_Type(xbm, T_ARRAY);
  data = rb_fltk_create_image(xbm);
  switch( argc ){
  case 3:
    image = new RBFLRGB_Image(data,NUM2INT(w),NUM2INT(h));
    break;
  case 4:
    image = new RBFLRGB_Image(data,NUM2INT(w),NUM2INT(h),NUM2INT(d));
    break;
  case 5:
    image = new RBFLRGB_Image(data,NUM2INT(w),NUM2INT(h),NUM2INT(d), NUM2INT(ld));
    break;
  }
  obj = rb_FLImage_new(klass,image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

VALUE rb_FLImage_destroy(VALUE self)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)image);
  if( image ){
    delete image;
  };

  return Qnil;
};

VALUE rb_FLRGB_Image_destroy(VALUE self)
{
  RBFLRGB_Image *image = (RBFLRGB_Image*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)image);
  if( image && image->array ){
    free((char*)(image->array));
    delete image;
  };

  return Qnil;
};

#endif

VALUE rb_FLImage_label(VALUE self, VALUE vwidget)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);
  RBFLWidget *widget = (RBFLWidget*)rb_to_fl(vwidget);

  image->label(widget);

  return Qnil;
};

VALUE rb_FLImage_draw(int argc, VALUE argv[], VALUE self)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);
  VALUE arg1, arg2, arg3, arg4, arg5, arg6;
  int x, y, w, h, cx, cy;

  switch( rb_scan_args(argc, argv, "24", &arg1,&arg2,&arg3,&arg4,&arg5,&arg6) ){
  case 2:
    x = NUM2INT(arg1);
    y = NUM2INT(arg2);
    image->draw(x,y);
    break;
  case 4:
    x = NUM2INT(arg1);
    y = NUM2INT(arg2);
    w = NUM2INT(arg3);
    h = NUM2INT(arg4);
    image->draw(x,y,w,h);
    break;
  case 6:
    x = NUM2INT(arg1);
    y = NUM2INT(arg2);
    w = NUM2INT(arg3);
    h = NUM2INT(arg4);
    cx = NUM2INT(arg5);
    cy = NUM2INT(arg6);
    image->draw(x,y,w,h,cx,cy);
    break;
  default:
    rb_raise(rb_eArgError, "too few/many arguments");
  };

  return Qnil;
};

extern "C" void
Init_flimage(){
  rb_cFLImage = rb_define_class_under(rb_mFLTK, "Image", rb_cObject);
  DEF_S_METHOD(Image, new, -1);
  DEF_METHOD(Image, initialize, -1);
  DEF_METHOD(Image, destroy, 0);
  DEF_METHOD(Image, label, 1);
  DEF_METHOD(Image, draw, -1);

#ifdef USE_RGB_IMAGE
  rb_cFLRGB_Image = rb_define_class_under(rb_mFLTK, "RGBImage", rb_cFLImage);
  DEF_S_METHOD(RGB_Image, new, -1);
  DEF_METHOD(RGB_Image, destroy, 0);
#endif
};
