/* -*- C++ -*-
 * $Id: flmenu.cc,v 1.17 2001/11/28 11:51:40 ttate Exp $
 */

#include "fltk.h"
#include "flmenu.h"

VALUE rb_cFLMenu_;
VALUE rb_cFLChoice;
VALUE rb_cFLMenu_Bar;
VALUE rb_cFLMenu_Button;

static FN_S_NEW(Menu_);
WIDGET_FNS(Menu_);
static FN_GETSET_INT(Menu_,value);
static FN_SET_INT2(Menu_,shortcut);
static FN_SET_INT_OPTINT(Menu_,mode);
static FN_GET_STR_SET_OPTINT(Menu_,text);
static FN_INIT2(Menu_, Fl_Widget::size, {
  rb_ivar_set(self, rb_id_iitems, rb_ary_new());
});

static void rb_flmenu_internal_func(Fl_Widget *widget, void *user_data)
{
  VALUE obj = rb_fltk_objmap_aref(widget);
  VALUE items = rb_ivar_get(obj, rb_id_iitems);
  VALUE proc = rb_ary_entry((VALUE)user_data, 0);
  VALUE data = rb_ary_entry((VALUE)user_data, 1);
  if( items == Qnil ) return ;
  if( proc == Qnil ) return ;
  rb_funcall(proc, rb_id_call, 2, obj, data);
};

static VALUE rb_FLMenu__add(int argc, VALUE argv[], VALUE self)
{
  RBFLMenu_ *widget = (RBFLMenu_*)rb_to_fl(self);
  VALUE label, shortcut, proc, data, flags;
  VALUE pass_data;

  if( argc == 1 ){
    char *text = STR2CSTR(argv[0]);
    int  len = strlen(text);
    int  n = 0;
    for( int i=0; i < len; i++ ){
      if( text[i] == '|' ){
	rb_ary_push(rb_ivar_get(self, rb_id_iitems),
		    rb_assoc_new(Qnil,Qnil));
      };
    };
    return INT2NUM(widget->add(text));
  };

  if( rb_block_given_p() ){
    proc = rb_f_lambda();
    switch( rb_scan_args(argc, argv, "22", &label, &shortcut, &data, &flags) ){
    case 2:
      pass_data = rb_assoc_new(proc, Qnil);
      flags = INT2NUM(0);
      break;
    case 3:
      pass_data = rb_assoc_new(proc, data);
      flags = INT2NUM(0);
      break;
    case 4:
      pass_data = rb_assoc_new(proc, data);
      break;
    };
  }
  else{
    switch( rb_scan_args(argc,argv,"32",&label,&shortcut,&proc,&data,&flags) ){
    case 3:
      pass_data = rb_assoc_new(proc, Qnil);
      flags = INT2NUM(0);
      break;
    case 4:
      pass_data = rb_assoc_new(proc, data);
      flags = INT2NUM(0);
      break;
    case 5:
      pass_data = rb_assoc_new(proc, data);
      break;
    };
  };

  rb_ary_push(rb_ivar_get(self, rb_id_iitems), pass_data);
  if( TYPE(shortcut) == T_STRING ){
    widget->add(STR2CSTR(label), STR2CSTR(shortcut),
		rb_flmenu_internal_func, (void*)pass_data,
		NUM2INT(flags));
  }
  else if( shortcut == Qnil ){
    widget->add(STR2CSTR(label), 0,
		rb_flmenu_internal_func, (void*)pass_data,
		NUM2INT(flags));
  }
  else{
    widget->add(STR2CSTR(label), NUM2INT(shortcut),
		rb_flmenu_internal_func, (void*)pass_data,
		NUM2INT(flags));
  };
  return Qnil;
};

static VALUE rb_FLMenu__remove(VALUE self, VALUE n)
{
  RBFLMenu_ *widget = (RBFLMenu_*)rb_to_fl(self);

  widget->remove(NUM2INT(n));
  rb_ary_delete_at(rb_ivar_get(self, rb_id_iitems), NUM2INT(n) - 1);
  return Qnil;
};

static VALUE rb_FLMenu__clear(VALUE self)
{
  RBFLMenu_ *widget = (RBFLMenu_*)rb_to_fl(self);
  widget->clear();
  rb_ary_clear(rb_ivar_get(self, rb_id_iitems));
  return Qnil;
};

static FN_S_NEW(Choice);
WIDGET_FNS(Choice);

static FN_S_NEW(Menu_Bar);
WIDGET_FNS(Menu_Bar);

static FN_S_NEW(Menu_Button);
WIDGET_FNS(Menu_Button);
FN_GET_WIDGET(Menu_Button,popup);
FN_SET_INT_RENAME(Menu_Button,menu_button_type,type);

extern "C" void
Init_flmenu(){
  rb_cFLMenu_ = rb_define_class_under(rb_mFLTK, "Menu", rb_cFLWidget);
  DEF_S_METHOD(Menu_,new,-1);
  DEF_WIDGET_METHODS(Menu_);
  DEF_METHOD(Menu_,initialize,-1);
  DEF_METHOD(Menu_,add,-1);
  DEF_METHOD(Menu_,remove,1);
  DEF_METHOD(Menu_,clear,0);
  DEF_METHOD(Menu_,value,-1);
  DEF_METHOD(Menu_,shortcut,2);
  DEF_METHOD(Menu_,mode,-1);
  DEF_METHOD(Menu_,text,-1);
#define DEF_MENUMODE(c) \
       rb_define_const(rb_mFLTK, #c, INT2NUM(FL_##c))
  DEF_MENUMODE(MENU_INACTIVE);
  DEF_MENUMODE(MENU_TOGGLE);
  DEF_MENUMODE(MENU_VALUE);
  DEF_MENUMODE(MENU_RADIO);
  DEF_MENUMODE(MENU_INVISIBLE);
  DEF_MENUMODE(SUBMENU_POINTER);
  DEF_MENUMODE(SUBMENU);
  DEF_MENUMODE(MENU_DIVIDER);
  DEF_MENUMODE(MENU_HORIZONTAL);
#undef DEF_MENUMODE


  rb_cFLChoice = rb_define_class_under(rb_mFLTK, "Choice", rb_cFLMenu_);
  DEF_S_METHOD(Choice,new,-1);
  DEF_WIDGET_METHODS(Choice);

  rb_cFLMenu_Bar = rb_define_class_under(rb_mFLTK, "MenuBar", rb_cFLMenu_);
  DEF_S_METHOD(Menu_Bar,new,-1);
  DEF_WIDGET_METHODS(Menu_Bar);

  rb_cFLMenu_Button = rb_define_class_under(rb_mFLTK, "MenuButton", rb_cFLMenu_);
  DEF_S_METHOD(Menu_Button,new,-1);
  DEF_WIDGET_METHODS(Menu_Button);
  DEF_METHOD(Menu_Button,popup,0);
  DEF_METHOD_EQ(Menu_Button,menu_button_type,1);
#define DEF_MENUBUTTON_CONST(c) \
       rb_define_const(rb_cFLMenu_Button, #c, INT2NUM(Fl_Menu_Button::c))
  DEF_MENUBUTTON_CONST(POPUP1);
  DEF_MENUBUTTON_CONST(POPUP2);
  DEF_MENUBUTTON_CONST(POPUP3);
  DEF_MENUBUTTON_CONST(POPUP12);
  DEF_MENUBUTTON_CONST(POPUP13);
  DEF_MENUBUTTON_CONST(POPUP23);
  DEF_MENUBUTTON_CONST(POPUP123);
#undef DEF_MENUBUTTON_CONST

};
