/* -*- C++ -*-
 * $Id: fltk.cc,v 1.75 2001/12/05 08:02:25 ttate Exp $
 */

#include "fltk.h"
#include "flenum.h"
#include "flwidget.h"
#include "flbox.h"
#include "flbutton.h"
#include "flbrowser.h"
#include "flchart.h"
#include "flclock.h"
#include "flgroup.h"
#include "flcolorchooser.h"
#include "flpack.h"
#include "flscroll.h"
#include "fltabs.h"
#include "fltile.h"
#include "flwindow.h"
#include "flinput.h"
#include "flmenu.h"
#include "flpositioner.h"
#include "flvaluator.h"
#include "flpixmap.h"
#include "flbitmap.h"
#include "flimage.h"

#if FL_MAJOR_VERSION >= 1 && FL_MINOR_VERSION >= 1
# include "flhelp.h"
# include "flfile.h"
#endif

#include <FL/fl_file_chooser.H>
#include <FL/fl_ask.H>

#if defined(USE_GL)
# include <FL/gl.h>
# include <FL/glut.h>
#endif

/* test */
#if defined(USE_XSETLOCALE)
#  include <X11/Xlocale.h>
#elif defined(USE_SETLOCALE)
#  include <locale.h>
#endif


VALUE rb_mFLTK;
VALUE rb_mFLTKGL;
VALUE rb_eFLError;
static VALUE rb_fltk_objmap;
static VALUE rb_fltk_fdmap;
static VALUE rb_fltk_callbacks;
static VALUE rb_fltk_handlers;
static double rb_fltk_thread_schedule_time = RBFL_THREAD_SCHEDULE_TIME;

ID rb_id_handle;
ID rb_id_phandle;
ID rb_id_draw;
ID rb_id_pdraw;
ID rb_id_icallback;
ID rb_id_icallback_data;
ID rb_id_iitems;
ID rb_id_call;
ID rb_id_clear;
ID rb_id_resize;
ID rb_id_flush;
ID rb_id_show;
ID rb_id_hide;
ID rb_id_draw_overlay;

#ifndef HAVE_RB_HASH_DELETE
extern "C" {
typedef struct st_table_entry {
  unsigned int hash;
  char *key;
  char *record;
  st_table_entry *next;
} st_table_entry;
extern int st_delete(st_table *table, char **key, char **value);
extern int st_delete_safe(st_table *table, char **key, char **value, char *never);
#define HASH_DELETED  FL_USER1
static void
rb_hash_modify(VALUE hash)
{
    if (OBJ_FROZEN(hash)) rb_error_frozen("hash");
    if (!OBJ_TAINTED(hash) && rb_safe_level() >= 4)
        rb_raise(rb_eSecurityError, "Insecure: can't modify hash");
};

static VALUE
rb_hash_delete(VALUE hash, VALUE key)
{
    VALUE val;

    rb_hash_modify(hash);
    if (RHASH(hash)->iter_lev > 0) {
        if (st_delete_safe(RHASH(hash)->tbl, (char**)(&key), (char**)(&val), (char*)(Qundef))) {
            FL_SET(hash, HASH_DELETED);
            return val;
        }
    }
    else if (st_delete(RHASH(hash)->tbl, (char**)(&key), (char**)(&val)))
        return val;
    if (rb_block_given_p()) {
        return rb_yield(key);
    }
    return RHASH(hash)->ifnone;
}
} /* extern "C" */
#undef HASH_DELETED
#endif

#ifndef HAVE_RB_HASH_CLEAR
static VALUE
rb_hash_clear(VALUE hash)
{
  rb_funcall(hash, rb_id_clear, 0);
  return hash;
}
#endif /* HAVE_RB_HASH_CLEAR */

void
rb_fltk_objmap_clear()
{
  rb_hash_clear(rb_fltk_objmap);
};

void
rb_fltk_objmap_delete(Fl_Widget *w)
{
  rb_hash_delete(rb_fltk_objmap, LONG2NUM((long)w));
};

void
rb_fltk_objmap_aset(Fl_Widget *w, VALUE obj)
{
  rb_hash_aset(rb_fltk_objmap, LONG2NUM((long)w), obj);
};

VALUE
rb_fltk_objmap_aref(Fl_Widget *w)
{
  return rb_hash_aref(rb_fltk_objmap, LONG2NUM((long)w));
};

int *iary2cary(VALUE v, int last)
{
  int i, len;
  int *ary;
  VALUE e;

  if( v == Qnil ) return NULL;

  Check_Type(v, T_ARRAY);
  len = RARRAY(v)->len;
  ary = (int*)malloc(sizeof(long) * (len+1));
  for( i=0; i < len; i++ ){
    e = rb_ary_entry(v, i);
    switch( TYPE(e) ){
    case T_FIXNUM:
    case T_BIGNUM:
      ary[i] = (int)(NUM2INT(e));
      break;
    case T_NIL:
      ary[i] = 0;
      break;
    default:
      rb_raise(rb_eTypeError, "unexpected type of the element #%d", i);
      break;
    };
  };
  ary[len] = last;

  return ary;
};

VALUE cary2iary(int *is, int last)
{
  VALUE ary = rb_ary_new();

  if( !is ) return Qnil;

  for( int i=0; is[i] != last; i++ ){
    rb_ary_push(ary, INT2NUM(is[i]));
  };

  return ary;
};

VALUE cary2iary_n(int *is, int len)
{
  VALUE ary = rb_ary_new();

  if( !is ) return Qnil;

  for( int i=0; i < len; i++ ){
    rb_ary_push(ary, INT2NUM(is[i]));
  };

  return ary;
};

float *fary2cary(VALUE v, float last)
{
  int i, len;
  float *ary;
  VALUE e;

  if( v == Qnil ) return NULL;

  Check_Type(v, T_ARRAY);
  len = RARRAY(v)->len + 1;
  ary = (float*)malloc(sizeof(double) * len);
  for( i=0; i < len - 1; i++ ){
    e = rb_ary_entry(v, i);
    switch( TYPE(e) ){
    case T_FLOAT:
      ary[i] = (float)(RFLOAT(e)->value);
      break;
    case T_NIL:
      ary[i] = 0.0;
      break;
    default:
      rb_raise(rb_eTypeError, "unexpected type of the element #%d", i);
      break;
    };
  };
  ary[len] = last;

  return ary;
};

VALUE cary2fary_n(float *fs, int len)
{
  VALUE ary = rb_ary_new();

  if( !fs ) return Qnil;

  for( int i=0; i < len; i++ ){
    rb_ary_push(ary, rb_float_new(fs[i]));
  };

  return ary;
};

char ** rb_fltk_create_pixmap(VALUE ary)
{
  int len;
  VALUE hd;
  long w, h, colors, chars_per_pixel, size;
  char ** data;

  Check_Type(ary, T_ARRAY);
  len = RARRAY(ary)->len;
  if( len <= 0 ) return NULL;

  data = (char**)malloc(sizeof(char*) * (len + 1));
  for( int i=0; i<=len-1; i++ ){
    VALUE str = RARRAY(ary)->ptr[i];
    Check_Type(str, T_STRING);
    data[i] = (char*)malloc(RSTRING(str)->len+1);
    memcpy(data[i], STR2CSTR(str), RSTRING(str)->len);
    data[i][RSTRING(str)->len] = '\0';
  };
  data[len] = NULL;

  return (char**)data;
};

char *rb_fltk_create_bitmap(VALUE ary)
{
  int len;
  VALUE hd;
  char *data;

  Check_Type(ary, T_ARRAY);
  len = RARRAY(ary)->len;
  data = (char*)malloc(len);
  for( int i=0; i<=len-1; i++ ){
    VALUE c = RARRAY(ary)->ptr[i];
    data[i] = (char)NUM2INT(c);
  };

  return data;
};

uchar *rb_fltk_create_image(VALUE ary)
{
  int len;
  VALUE hd;
  uchar *data;

  Check_Type(ary, T_ARRAY);
  len = RARRAY(ary)->len;
  data = (uchar*)malloc(len);
  for( int i=0; i<=len-1; i++ ){
    VALUE c = RARRAY(ary)->ptr[i];
    data[i] = (uchar)NUM2UINT(c);
  };

  return data;
};


static void
rb_fltk_call_thread_schedule(void *data)
{
  if( rb_fltk_thread_schedule_time > 0.0 ){
    rb_thread_schedule();
    Fl::repeat_timeout(rb_fltk_thread_schedule_time, rb_fltk_call_thread_schedule);
  };
};

static VALUE
rb_fltk_enable_thread(int argc, VALUE argv[], VALUE self)
{
  VALUE val;
  int add = 0;

  if( rb_fltk_thread_schedule_time <= 0.0 ) add = 1;

  switch( rb_scan_args(argc, argv, "01", &val) ){
  case 0:
    rb_fltk_thread_schedule_time = RBFL_THREAD_SCHEDULE_TIME;
    break;
  case 1:
    Check_Type(val, T_FLOAT);
    rb_fltk_thread_schedule_time = RFLOAT(val)->value;
    break;
  };

  if( add ){
    Fl::add_timeout(rb_fltk_thread_schedule_time, rb_fltk_call_thread_schedule);
  };

  return Qnil;
};

static VALUE
rb_fltk_disable_thread(VALUE self)
{
  rb_fltk_thread_schedule_time = -1.0;
  return Qnil;
};

static void
rb_fltk_add_fd_func(int fd, void *data)
{
  VALUE proc;

  proc = rb_hash_aref(rb_fltk_fdmap, INT2NUM(fd));
  rb_funcall(proc, rb_id_call, 2, INT2NUM(fd), (VALUE)data);
};

static VALUE rb_fltk_add_fd(int argc, VALUE argv[], VALUE mod)
{
  VALUE arg1, arg2, arg3, arg4;
  int fd = -1;
  int when;
  VALUE data, proc;

  if( rb_block_given_p() ){
    switch( rb_scan_args(argc, argv, "12", &arg1, &arg2, &arg3) ){
    case 1:
      fd = NUM2INT(rb_Integer(arg1));
      proc = rb_f_lambda();
      data = Qnil;
      Fl::add_fd(fd, rb_fltk_add_fd_func, (void*)data);
      break;
    case 2:
      fd = NUM2INT(rb_Integer(arg1));
      proc = rb_f_lambda();
      data = arg2;
      Fl::add_fd(fd, rb_fltk_add_fd_func, (void*)data);
      break;
    case 3:
      fd = NUM2INT(rb_Integer(arg1));
      when = NUM2INT(arg2);
      proc = rb_f_lambda();
      data = arg3;
      Fl::add_fd(fd, when, rb_fltk_add_fd_func, (void*)data);
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "22", &arg1, &arg2, &arg3, &arg4) ){
    case 2:
      fd = NUM2INT(rb_Integer(arg1));
      proc = arg2;
      data = Qnil;
      Fl::add_fd(fd, rb_fltk_add_fd_func, (void*)data);
      break;
    case 3:
      fd = NUM2INT(rb_Integer(arg1));
      proc = arg2;
      data = arg3;
      Fl::add_fd(fd, rb_fltk_add_fd_func, (void*)data);
      break;
    case 4:
      fd = NUM2INT(rb_Integer(arg1));
      when = NUM2INT(arg2);
      proc = arg3;
      data = arg4;
      Fl::add_fd(fd, when, rb_fltk_add_fd_func, (void*)data);
      break;
    };
  };

  rb_hash_aset(rb_fltk_fdmap, INT2NUM(fd), proc);

  return rb_ary_new3(3, INT2NUM(fd), proc, data);
};

static VALUE rb_fltk_remove_fd(VALUE mod, VALUE fd)
{
  Fl::remove_fd(NUM2INT(fd));
  rb_hash_delete(rb_fltk_fdmap, fd);
  return Qnil;
};

static void rb_fltk_internal_callback(void *user_data)
{
  VALUE proc = rb_ary_entry((VALUE)user_data, 0);
  VALUE data = rb_ary_entry((VALUE)user_data, 1);

  if( rb_ary_includes(rb_fltk_callbacks, (VALUE)user_data) == Qtrue ){
    rb_funcall(proc, rb_id_call, 1, data);
  };
};

static void rb_fltk_internal_timeout(void *user_data)
{
  VALUE proc = rb_ary_entry((VALUE)user_data, 0);
  VALUE data = rb_ary_entry((VALUE)user_data, 1);

  if( rb_ary_includes(rb_fltk_callbacks, (VALUE)user_data) == Qtrue ){
    rb_funcall(proc, rb_id_call, 1, data);
    /* the below line is not appeared in rb_fltk_internal_callback() */
    rb_ary_delete(rb_fltk_callbacks, (VALUE)user_data);
  };
};

static VALUE rb_fltk_add_idle(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc;
  VALUE data;
  VALUE user_data;

  if( rb_block_given_p() ){
    proc = rb_f_lambda();
    switch( rb_scan_args(argc, argv, "01", &data) ){
    case 0:
      data = Qnil;
      break;
    case 1:
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "11", &proc, &data) ){
    case 1:
      data = Qnil;
      break;
    case 2:
      break;
    };
  };

  user_data = rb_assoc_new(proc, data);
  Fl::add_idle(rb_fltk_internal_callback, (void*)user_data);
  rb_ary_push(rb_fltk_callbacks, user_data);
  return user_data;
};

static VALUE rb_fltk_remove_idle(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc;
  VALUE data;
  VALUE user_data;

  switch( rb_scan_args(argc, argv, "11", &proc, &data) ){
  case 1:
    data = Qnil;
    break;
  case 2:
    break;
  };
  user_data = rb_assoc_new(proc, data);
  Fl::remove_idle(rb_fltk_internal_callback, (void*)user_data);
  rb_ary_delete(rb_fltk_callbacks, user_data);
  return Qnil;
};

static VALUE rb_fltk_add_timeout(int argc, VALUE argv[], VALUE mod)
{
  VALUE time;
  VALUE proc;
  VALUE data;
  VALUE user_data;

  if( rb_block_given_p() ){
    proc = rb_f_lambda();
    switch( rb_scan_args(argc, argv, "11", &time, &data) ){
    case 1:
      data = Qnil;
      break;
    case 2:
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "21", &time, &proc, &data) ){
    case 2:
      data = Qnil;
      break;
    case 3:
      break;
    };
  };

  time = rb_Float(time);

  user_data = rb_assoc_new(proc, data);
  Fl::add_timeout(RFLOAT(time)->value, rb_fltk_internal_timeout, (void*)user_data);
  rb_ary_push(rb_fltk_callbacks, user_data);
  return user_data;
};

static VALUE rb_fltk_repeat_timeout(int argc, VALUE argv[], VALUE mod)
{
  VALUE time;
  VALUE proc;
  VALUE data;
  VALUE user_data;

  if( rb_block_given_p() ){
    proc = rb_f_lambda();
    switch( rb_scan_args(argc, argv, "11", &time, &data) ){
    case 1:
      data = Qnil;
      break;
    case 2:
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "21", &time, &proc, &data) ){
    case 2:
      data = Qnil;
      break;
    case 3:
      break;
    };
  };

  time = rb_Float(time);

  user_data = rb_assoc_new(proc, data);
  Fl::repeat_timeout(RFLOAT(time)->value, rb_fltk_internal_timeout, (void*)user_data);
  rb_ary_push(rb_fltk_callbacks, user_data);
  return user_data;
};

static int rb_fltk_internal_handler(int e)
{
  int   len = RARRAY(rb_fltk_handlers)->len;
  VALUE *ptr    = RARRAY(rb_fltk_handlers)->ptr;
  VALUE proc;

  for( int i = 0; i < len; i++ ){
    proc = ptr[i];
    if( rb_funcall(proc, rb_id_call, 1, INT2NUM(e)) == Qtrue ){
      return 1;
    };
  };
  return 0;
};

static VALUE rb_fltk_add_handler(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc;

  if( rb_block_given_p() ){
    if( argc == 0 ){
      proc = rb_f_lambda();
    }
    else{
      rb_raise(rb_eArgError, "too many arguments");
    };
  }
  else{
    if( argc == 1 ){
      proc = argv[0];
    }
    else{
      rb_raise(rb_eArgError, "too few/many arguments");
    };
  };

  rb_ary_push(rb_fltk_handlers, proc);

  return Qnil;
};

static VALUE rb_fltk_draw(int argc, VALUE argv[], VALUE self)
{
  VALUE str, x, y, n, w, h, align;

  switch( rb_scan_args(argc, argv, "33", &str, &x, &y, &w, &h, &align) ){
  case 3:
    fl_draw((const char *)STR2CSTR(str), NUM2INT(x), NUM2INT(y));
    break;
  case 4:
    n = x; x = y; y = w;
    fl_draw((const char *)STR2CSTR(str), NUM2INT(n),
	    NUM2INT(x), NUM2INT(y));
    break;
  case 5:
    fl_draw((const char *)STR2CSTR(str), NUM2INT(x), NUM2INT(y),
	    NUM2INT(w), NUM2INT(h), FL_ALIGN_CENTER);
    break;
  case 6:
    fl_draw((const char *)STR2CSTR(str), NUM2INT(x), NUM2INT(y),
	    NUM2INT(w), NUM2INT(h), (Fl_Align)NUM2INT(align));
    break;
  };
  return Qnil;
};

static VALUE rb_fltk_width(int argc, VALUE argv[], VALUE self)
{
  VALUE str, n;
  double ret;

  switch( rb_scan_args(argc, argv, "11", &str, &n) ){
  case 1:
    ret = fl_width(STR2CSTR(str));
    break;
  case 2:
    ret = fl_width(STR2CSTR(str), NUM2INT(n));
    break;
  };
  return rb_float_new(ret);
};

static VALUE rb_fltk_clip_box(VALUE mod, VALUE x, VALUE y, VALUE w, VALUE h)
{
  int rx,ry,rw,rh;

  fl_clip_box(NUM2INT(x), NUM2INT(y), NUM2INT(w), NUM2INT(h),
	      rx, ry, rw, rh);
  return rb_ary_new3(4, INT2NUM(rx), INT2NUM(ry), INT2NUM(rw), INT2NUM(rh));
};

static VALUE rb_fltk_line_style(VALUE argc, VALUE argv[], VALUE self)
{
    VALUE arg1, arg2, arg3;
    char *ary;
    int len;

    switch( rb_scan_args(argc, argv, "12", &arg1, &arg2, &arg3) ){
      case 1:
        fl_line_style(NUM2INT(arg1));
	break;
      case 2:
        fl_line_style(NUM2INT(arg1), NUM2INT(arg2));
	break;
      case 3:
	if( arg3 == Qnil ){
	  ary = 0;
	}
	else{
	  Check_Type(arg3, T_ARRAY);
	  len = RARRAY(arg3)->len;
	  ary = ALLOCA_N(char, len + 1);
	  for( int i = 0; i < len; i++ ){
	    ary[i] = RARRAY(arg3)->ptr[i];
	  };
	  ary[len] = 0;
	};
	fl_line_style(NUM2INT(arg1), NUM2INT(arg2), ary);
        break;
    };
    return Qnil;
};

static VALUE rb_fltk_set_font(VALUE self, VALUE arg1, VALUE arg2)
{
  Check_Type(arg2,T_STRING);
  Fl::set_font((Fl_Font)NUM2INT(arg1), STR2CSTR(arg2));

  return Qnil;
};

static VALUE rb_fltk_set_fonts(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1;
  int ret;
  switch( rb_scan_args(argc, argv, "01", &arg1) ){
  case 0:
    ret = Fl::set_fonts();
    break;
  case 1:
    ret = Fl::set_fonts(STR2CPTR(arg1));
    break;
  };
  return INT2NUM(ret);
};

static VALUE rb_fltk_event_key(int argc, VALUE argv[], VALUE self)
{
  VALUE ch;

  switch( rb_scan_args(argc, argv, "01", &ch) ){
  case 0:
    return INT2NUM(Fl::event_key());
  case 1:
    return INT2BOOL(Fl::event_key((int)NUM2INT(ch)));
  };
};

static VALUE rb_fltk_size(int argc, VALUE argv[], VALUE self)
{
  VALUE size;

  switch( rb_scan_args(argc, argv, "01", &size) ){
  case 0:
    size = INT2NUM(fl_size());
    return size;
  case 1:
    fl_font(fl_font(), NUM2INT(size));
    return Qnil;
  };
};

static VALUE rb_fltk_font(int argc, VALUE argv[], VALUE self)
{
  VALUE font, size;

  switch( rb_scan_args(argc, argv, "02", &font, &size) ){
  case 0:
    font = INT2NUM(fl_font());
    return font;
  case 1:
    fl_font(NUM2INT(font), fl_size());
    return Qnil;
  case 2:
    fl_font(NUM2INT(font), NUM2INT(size));
    return Qnil;
  };
};

static VALUE rb_fltk_run(VALUE self)
{
  Fl::run();
  rb_fltk_objmap_clear();
  return Qnil;
};

static FN_MODULE_FL_CALL(check);
static FN_MODULE_FL_GET_INT(event_x);
static FN_MODULE_FL_GET_INT(event_y);
static FN_MODULE_FL_GET_INT(event_button);
static FN_MODULE_FL_GET_INT_SET_INT(get_key);
static FN_GLOBAL_GETSET_INT(color);
static FN_GLOBAL_SET_INT4_OPTINT2(line);
static FN_GLOBAL_SET_INT4(push_clip);
static FN_GLOBAL_CALL(push_no_clip);
static FN_GLOBAL_CALL(pop_clip);
static FN_GLOBAL_GET_BOOL_SET_INT4(not_clipped);
static FN_GLOBAL_SET_INT4(rect);
static FN_GLOBAL_SET_INT4_OPTCHR3(rectf);
static FN_GLOBAL_SET_INT4_DBL2(pie);
static FN_GLOBAL_SET_INT4_DBL2(arc);
static FN_GLOBAL_SET_DBL3(circle);
static FN_GLOBAL_SET_DBL2(vertex);
static FN_GLOBAL_CALL(begin_line);
static FN_GLOBAL_CALL(end_line);
static FN_GLOBAL_CALL(begin_loop);
static FN_GLOBAL_CALL(end_loop);
static FN_GLOBAL_CALL(begin_polygon);
static FN_GLOBAL_CALL(end_polygon);
static FN_GLOBAL_CALL(begin_complex_polygon);
static FN_GLOBAL_CALL(end_complex_polygon);
static FN_GLOBAL_CALL(gap);
static FN_GLOBAL_SET_INT2(point);
static FN_GLOBAL_SET_FORMAT(message);
static FN_GLOBAL_SET_FORMAT(alert);
static FN_GLOBAL_GET_BOOL_SET_FORMAT(ask);
static FN_GLOBAL_GET_INT_SET_FORMAT_STR3(choice);
static FN_GLOBAL_GET_STR_SET_FORMAT_STR(input);
static FN_GLOBAL_GET_STR_SET_FORMAT_STR(password);

static VALUE rb_fltk_wait(int argc, VALUE argv[], VALUE self)
{
VALUE time;
  switch(rb_scan_args(argc,argv,"01",&time))
  {
    case 0:
      return INT2NUM(Fl::wait());
    case 1:
      return rb_float_new(Fl::wait());
  }
  return Qnil;
}

static VALUE rb_fltk_draw_pixmap(int argc, VALUE argv[], VALUE self)
{
  VALUE xpm, x, y, color;
  int r;
  char **data;

  rb_scan_args(argc, argv, "31", &xpm, &x, &y, &color);
  data = rb_fltk_create_pixmap(xpm);
  switch( argc ){
  case 3:
    r = INT2NUM(fl_draw_pixmap(data, NUM2INT(x), NUM2INT(y)));
    break;
  case 4:
    r = INT2NUM(fl_draw_pixmap(data, NUM2INT(x), NUM2INT(y),
			       (Fl_Color)NUM2INT(color)));
    break;
  };
  return r;
};

static VALUE rb_fltk_measure_pixmap(VALUE self, VALUE xpm)
{
  VALUE str;
  char *data[2];
  int r, w, h;

  Check_Type(xpm, T_ARRAY);
  str = RARRAY(xpm)->ptr[0];
  data[0] = STR2CSTR(str);
  data[1] = NULL;
  r = fl_measure_pixmap(data, w, h);
  if( r ){
    return rb_ary_new3(2, INT2NUM(w), INT2NUM(h));
  }
  else{
    return Qnil;
  };
};

static VALUE rb_fltk_file_chooser_callback_obj = Qnil;

static void rb_fltk_file_chooser_callback_internal(const char *str)
{
  if( rb_fltk_file_chooser_callback_obj != Qnil ){
    rb_funcall(rb_fltk_file_chooser_callback_obj, rb_id_call, 1, CPTR2STR(str));
  };
};

static VALUE rb_fltk_file_chooser_callback(int argc, VALUE argv[], VALUE self)
{
  VALUE proc;

  switch( rb_scan_args(argc, argv, "01", proc) ){
  case 0:
    proc = rb_f_lambda();
    break;
  case 1:
    /* do nothing */
    break;
  };
  rb_fltk_file_chooser_callback_obj = proc;
  fl_file_chooser_callback(rb_fltk_file_chooser_callback_internal);

  return Qnil;
};

static VALUE rb_fltk_file_chooser(int argc, VALUE argv[], VALUE self)
{
  VALUE msg, pat, fname;
  const char *c_msg, *c_pat, *c_fname;
  char *ret;

  switch( rb_scan_args(argc, argv, "12", &msg, &pat, &fname) ){
  case 1:
    c_msg = STR2CPTR(msg);
    c_pat = NULL;
    c_fname = NULL;
    break;
  case 2:
    c_msg = STR2CPTR(msg);
    c_pat = STR2CPTR(pat);
    c_fname = NULL;
    break;
  case 3:
    c_msg = STR2CPTR(msg);
    c_pat = STR2CPTR(pat);
    c_fname = STR2CPTR(fname);
    break;
  };

  ret = fl_file_chooser(c_msg, c_pat, c_fname);
  return ret ? rb_tainted_str_new2(ret) : Qnil;
};

static VALUE rb_fltk_paste( VALUE self, VALUE arg1 ){
  Fl_Widget *widget;

  widget =  rb_to_fl(arg1);
  if( widget ){
    if(Fl::check()){
      Fl::paste(*widget);
    }
    else{
      rb_raise(rb_eSystemCallError, "FLTK does not appear to be taking events!");
    };
  }
  else{
    rb_raise(rb_eTypeError, "Need widget to send paste-event to.");
  };
  return Qnil;
};

#ifdef USE_UTF8
#include <FL/fl_utf8.h>
static VALUE rb_fltk_utflen(VALUE self, VALUE str)
{
  Check_Type(str, T_STRING);
  return INT2NUM(fl_utflen((const unsigned char *)STR2CSTR(str),
			   RSTRING(str)->len));
};
static VALUE rb_fltk_utf_nb_char(VALUE self, VALUE str)
{
  Check_Type(str, T_STRING);
  return INT2NUM(fl_utflen((const unsigned char *)STR2CSTR(str),
			   RSTRING(str)->len));
};
#endif

#ifdef USE_GL
static FN_GL_GLOBAL_SET_ENUM(color, Fl_Color);
#endif

extern "C" void
Init_rubyfltk(void)
{
  rb_id_handle = rb_intern("handle");
  rb_id_draw   = rb_intern("draw");
  rb_id_phandle = rb_intern("_handle");
  rb_id_pdraw   = rb_intern("_draw");
  rb_id_icallback = rb_intern("@callback");
  rb_id_icallback_data = rb_intern("@callback_data");
  rb_id_iitems = rb_intern("@items");
  rb_id_call = rb_intern("call");
  rb_id_clear = rb_intern("clear");
  rb_id_resize = rb_intern("resize");
  rb_id_flush  = rb_intern("flush");
  rb_id_show   = rb_intern("show");
  rb_id_hide   = rb_intern("hide");
  rb_id_draw_overlay = rb_intern("draw_overlay");

  Fl::add_timeout(RBFL_THREAD_SCHEDULE_TIME, rb_fltk_call_thread_schedule);

  rb_mFLTK = rb_define_module("FLTK");
  rb_define_const(rb_mKernel, "Fltk", rb_mFLTK);

  rb_eFLError = rb_define_class_under(rb_mFLTK, "FLError", rb_eStandardError);

  rb_fltk_objmap = rb_hash_new();
  rb_fltk_fdmap  = rb_hash_new();
  rb_fltk_callbacks = rb_ary_new();
  rb_fltk_handlers  = rb_ary_new();
  rb_define_const(rb_mFLTK, "ObjectMap", rb_fltk_objmap);
  rb_define_const(rb_mFLTK, "FDMap", rb_fltk_fdmap);
  rb_define_const(rb_mFLTK, "Callbacks", rb_fltk_callbacks);
  rb_define_const(rb_mFLTK, "Handlers", rb_fltk_handlers);

  Fl::add_handler(rb_fltk_internal_handler);

  Init_flenum();
  Init_flwidget();
  Init_flbox();
  Init_flbutton();
  Init_flbrowser();
  Init_flchart();
  Init_flclock();
  Init_flgroup();
  Init_flcolorchooser();
  Init_flpack();
  Init_flscroll();
  Init_fltabs();
  Init_fltile();
  Init_flwindow();
  Init_flinput();
  Init_flmenu();
  Init_flpositioner();
  Init_flvaluator();
  Init_flpixmap();
  Init_flbitmap();
  Init_flimage();

#if FL_MAJOR_VERSION >= 1 && FL_MINOR_VERSION >= 1
  Init_flhelp();
  Init_flfile();
#endif

  DEF_MODULE_FUNC(enable_thread, -1);
  DEF_MODULE_FUNC(disable_thread, 0);
  DEF_MODULE_FUNC(add_fd, -1);
  DEF_MODULE_FUNC(remove_fd, 1);
  DEF_MODULE_FUNC(add_idle, -1);
  DEF_MODULE_FUNC(remove_idle, -1);
  DEF_MODULE_FUNC(add_timeout, -1);
  DEF_MODULE_FUNC(repeat_timeout, -1);
  DEF_MODULE_FUNC(add_handler, -1);
  DEF_MODULE_FUNC(run,0);
  DEF_MODULE_FUNC(check,0);
  DEF_MODULE_FUNC(set_font,2);
  DEF_MODULE_FUNC(set_fonts,-1);
  DEF_MODULE_FUNC(event_x,0);
  DEF_MODULE_FUNC(event_y,0);
  DEF_MODULE_FUNC(event_button,0);
  DEF_MODULE_FUNC(event_key,-1);
  DEF_MODULE_FUNC(get_key,1);

  DEF_MODULE_FUNC(push_clip,4);
  DEF_MODULE_FUNC(push_no_clip,0);
  DEF_MODULE_FUNC(pop_clip,0);
  DEF_MODULE_FUNC_Q(not_clipped,4);
  DEF_MODULE_FUNC(clip_box,4);
  DEF_MODULE_FUNC(paste,1);
  DEF_MODULE_FUNC_EQ(line_style,-1);
  DEF_MODULE_FUNC(draw,-1);
  DEF_MODULE_FUNC_EQ(font,-1);
  DEF_MODULE_FUNC_EQ(size,-1);
  DEF_MODULE_FUNC(width,-1);
  DEF_MODULE_FUNC_EQ(color,-1);
  DEF_MODULE_FUNC(line,-1);
  DEF_MODULE_FUNC(rect,4);
  DEF_MODULE_FUNC(rectf,-1);
  DEF_MODULE_FUNC(pie,6);
  DEF_MODULE_FUNC(arc,6);
  DEF_MODULE_FUNC(point,2);
  DEF_MODULE_FUNC(circle,3);
  DEF_MODULE_FUNC(begin_loop,0);
  DEF_MODULE_FUNC(end_loop,0);
  DEF_MODULE_FUNC(begin_line,0);
  DEF_MODULE_FUNC(end_line,0);
  DEF_MODULE_FUNC(begin_polygon,0);
  DEF_MODULE_FUNC(end_polygon,0);
  DEF_MODULE_FUNC(begin_complex_polygon,0);
  DEF_MODULE_FUNC(end_complex_polygon,0);
  DEF_MODULE_FUNC(gap,0);
  DEF_MODULE_FUNC(vertex,2);
  DEF_MODULE_FUNC(draw_pixmap,-1);
  DEF_MODULE_FUNC(measure_pixmap,1);

  DEF_MODULE_FUNC(message,-1);
  DEF_MODULE_FUNC(alert,-1);
  DEF_MODULE_FUNC(ask,-1);
  DEF_MODULE_FUNC(choice,-1);
  DEF_MODULE_FUNC(input,-1);
  DEF_MODULE_FUNC(password,-1);

  DEF_MODULE_FUNC(file_chooser,-1);
  DEF_MODULE_FUNC(file_chooser_callback,-1);
  DEF_MODULE_FUNC(wait,-1);

#ifdef USE_GL
  rb_mFLTKGL = rb_define_module_under(rb_mFLTK, "GL");
  DEF_GL_MODULE_FUNC(color,1);
#endif
#ifdef USE_UTF8
  rb_define_const(rb_mFLTK, "UTF8", Qtrue);
  DEF_MODULE_FUNC(utflen,1);
  DEF_MODULE_FUNC(utf_nb_char,1);
#endif
#if defined(USE_SETLOCALE) || defined(USE_XSETLOCALE)
  /* here is a test code */
  if( !setlocale(LC_CTYPE, getenv("LANG")) ){
    if( !setlocale(LC_CTYPE, "") ){
      /* raise error? */
    };
  };
#endif
};
