/* -*- C++ -*- */

#ifndef RBFLTK_H
#define RBFLTK_H

#include <ruby.h>
#include <FL/Fl.H>
#include <FL/fl_draw.H>

#ifndef LONG2NUM
# define LONG2NUM(x) INT2NUM(x)
#endif

#define STR2CPTR(x) ((x == Qnil) ? NULL : STR2CSTR(x))
#define INT2BOOL(x) (x ? Qtrue : Qfalse)
#define BOOL2INT(x) (x == Qtrue ? 1 : 0)
#define CPTR2STR(x) (x ? rb_tainted_str_new2(x) : Qnil)

/* old RUBY_METHOD_FUNC() definition doesn't match to prototypes in those days. */
#ifndef ANYARGS
#undef  RUBY_METHOD_FUNC
#define RUBY_METHOD_FUNC(func) ((VALUE (*)())func)
#endif

#define RBFL_THREAD_SCHEDULE_TIME 0.1

extern VALUE rb_mFLTK;
extern VALUE rb_mFLTKGL;
extern VALUE rb_eFLError;

extern ID rb_id_handle;
extern ID rb_id_phandle;
extern ID rb_id_draw;
extern ID rb_id_pdraw;
extern ID rb_id_call;
extern ID rb_id_icallback;
extern ID rb_id_icallback_data;
extern ID rb_id_iitems;
extern ID rb_id_call;
extern ID rb_id_clear;
extern ID rb_id_resize;
extern ID rb_id_flush;
extern ID rb_id_show;
extern ID rb_id_hide;
extern ID rb_id_draw_overlay;

extern VALUE rb_fltk_objmap_aref(Fl_Widget *);
extern void rb_fltk_objmap_aset(Fl_Widget *, VALUE);
extern void rb_fltk_objmap_delete(Fl_Widget *);
extern int *iary2cary(VALUE val, int last = 0);
extern float *fary2cary(VALUE val, float last = 0.0);
extern VALUE cary2iary(int *is, int last = 0);
extern VALUE cary2iary_n(int *is, int len);
extern VALUE cary2fary_n(float *fs, int len);

#define RBFL_CALLER_HANDLE(Fl_Class) \
  virtual int handle(int e){ \
    extern VALUE rb_fl_destroyed(VALUE); \
    VALUE obj, ret, rb_object; \
    rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      ret = rb_funcall(rb_object, rb_id_handle, 1, INT2NUM(e)); \
      if( rb_fl_destroyed(rb_object) == Qtrue ) return 1; \
      switch( TYPE(ret) ){ \
      case T_FIXNUM: \
	return FIX2INT(ret); \
      case T_TRUE: \
	return 1; \
      default: \
	return 0; \
      }; \
    } \
    else{ \
      return 0; \
    }; \
  }; \
  virtual int call_parent_handle(int e){ \
    return Fl_Class::handle(e); \
  }; \

#define RBFL_CALLER_DRAW(Fl_Class) \
  virtual void draw(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_draw, 0); \
    }; \
  }; \
  virtual void call_parent_draw(){ \
    Fl_Class::draw(); \
  }; \

#define RBFL_CALLER_RESIZE(Fl_Class) \
  virtual void resize(int x, int y, int w, int h){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_resize, 4, \
            INT2NUM(x), INT2NUM(y), INT2NUM(w), INT2NUM(h)); \
    }; \
  }; \
  virtual void call_parent_resize(int x, int y, int w, int h){ \
    Fl_Class::resize(x,y,w,h); \
  };

#define RBFL_CALLER_FLUSH(Fl_Class) \
  virtual void flush(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_flush, 0); \
    }; \
  }; \
  virtual void call_parent_flush(){ \
    Fl_Class::flush(); \
  };

#define RBFL_CALLER_SHOW(Fl_Class) \
  virtual void show(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_show, 0); \
    }; \
  }; \
  virtual void call_parent_show(){ \
    Fl_Class::show(); \
  };

#define RBFL_CALLER_HIDE(Fl_Class) \
  virtual void hide(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_hide, 0); \
    }; \
  }; \
  virtual void call_parent_hide(){ \
    Fl_Class::hide(); \
  };

#define RBFL_CALLER_DRAW_OVERLAY(Fl_Class) \
  virtual void draw_overlay(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_draw_overlay, 0); \
    }; \
  }; \
  virtual void call_parent_draw_overlay(){ /* empty */ };

#define RBFL_CALLERS(Fl_Class) \
  RBFL_CALLER_HANDLE(Fl_Class) \
  RBFL_CALLER_DRAW(Fl_Class) \
  RBFL_CALLER_RESIZE(Fl_Class)

#define RBFL_HELP_CALLERS(Fl_Class) \
  RBFL_CALLER_RESIZE(Fl_Class) \

#define RBFL_WIDGET_CALLERS(Fl_Class) \
  RBFL_CALLER_HANDLE(Fl_Class) \
  virtual void draw(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_draw, 0); \
    }; \
  }; \
  virtual void call_parent_draw(){ /** do nothing **/ }; \
  RBFL_CALLER_RESIZE(Fl_Class) \

#define RBFL_WINDOW_CALLERS(Fl_Class) \
  RBFL_CALLERS(Fl_Class) \
  RBFL_CALLER_FLUSH(Fl_Class) \
  RBFL_CALLER_SHOW(Fl_Class) \
  RBFL_CALLER_HIDE(Fl_Class)

#define RBFL_OVERLAY_WINDOW_CALLERS(Fl_Class) \
  RBFL_WINDOW_CALLERS(Fl_Class) \
  RBFL_CALLER_DRAW_OVERLAY(Fl_Class)


#define RBFL_CLASS(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_HELP(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_HELP_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_WIDGET(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_WIDGET_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_WIN(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL##klass(int w, int h, const char *label = 0) \
    : Fl_##klass(w,h,label){}; \
  RBFL_WINDOW_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_OVERLAY_WIN(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL##klass(int w, int h, const char *label = 0) \
    : Fl_##klass(w,h,label){}; \
  RBFL_OVERLAY_WINDOW_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_OVERLAY_GLWIN(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL##klass(int w, int h, const char *label = 0) \
    : Fl_##klass(w,h,label){}; \
  RBFL_OVERLAY_WINDOW_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_TIMER(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(unsigned char c, int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(c,x,y,w,h,label){}; \
  RBFL_CALLERS(Fl_##klass); \
};

/*
 * naming policy:
 * rb_FLWidget_method(...) for class FLWidget and Fl_Widget
 * rb_fltk_method(...)   for module Fltk
 */

#define FN_S_NEW(klass) \
VALUE rb_FL##klass##_s_new(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  widget = new RBFL##klass(0, 0, 0, 0); \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_S_NEW_WIN(klass) \
VALUE rb_FL##klass##_s_new(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  widget = new RBFL##klass(0, 0); \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_S_POPUP_WIN(klass) \
VALUE rb_FL##klass##_s_popup(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  widget = new RBFL##klass(0, 0, 0, 0); \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_S_GET_BOOL_SET_INT(klass,method) \
VALUE rb_FL##klass##_s_##method(VALUE self, VALUE arg1) \
{ \
  return INT2BOOL(RBFL##klass::method(NUM2INT(arg1))); \
};

#define FN_INIT(klass) \
VALUE rb_FL##klass##_initialize(int argc, VALUE argv[], VALUE self) \
{ \
  VALUE obj; \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  VALUE rx,ry,rw,rh,rlabel; \
  switch( rb_scan_args(argc, argv, "05", &rx, &ry, &rw, &rh, &rlabel) ){ \
  case 0: \
    break; \
  case 1: \
    rlabel = rx; \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 2: \
    rw = rx; rh = ry; \
    rx = INT2NUM(widget->x()); ry = INT2NUM(widget->y()); \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    break; \
  case 3: \
    rlabel = rw; rw = rx; rh = ry; \
    rx = INT2NUM(widget->x()); ry = INT2NUM(widget->y()); \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 4: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    break; \
  case 5: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  }; \
  return Qnil; \
};

#define FN_GROUP_INIT(klass) \
VALUE rb_FL##klass##_initialize(int argc, VALUE argv[], VALUE self) \
{ \
  VALUE obj; \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  VALUE rx,ry,rw,rh,rlabel; \
  switch( rb_scan_args(argc, argv, "05", &rx, &ry, &rw, &rh, &rlabel) ){ \
  case 0: \
    break; \
  case 1: \
    rlabel = rx; \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 2: \
    rw = rx; rh = ry; \
    widget->size(NUM2INT(rw), NUM2INT(rh)); \
    break; \
  case 3: \
    rlabel = rw; rw = rx; rh = ry; \
    widget->size(NUM2INT(rw), NUM2INT(rh)); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 4: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    break; \
  case 5: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  }; \
  if( rb_block_given_p() ){ \
    rb_ensure(RUBY_METHOD_FUNC(rb_yield), self, \
              RUBY_METHOD_FUNC(rb_FLGroup_end), self); \
  }; \
  return Qnil; \
};

#define FN_INIT2(klass, size, extcode) \
VALUE rb_FL##klass##_initialize(int argc, VALUE argv[], VALUE self) \
{ \
  VALUE obj; \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  VALUE rx,ry,rw,rh,rlabel; \
  switch( rb_scan_args(argc, argv, "05", &rx, &ry, &rw, &rh, &rlabel) ){ \
  case 0: \
    break; \
  case 1: \
    rlabel = rx; \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 2: \
    rw = rx; rh = ry; \
    widget->size(NUM2INT(rw), NUM2INT(rh)); \
    break; \
  case 3: \
    rlabel = rw; rw = rx; rh = ry; \
    widget->size(NUM2INT(rw), NUM2INT(rh)); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 4: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    break; \
  case 5: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  }; \
  extcode; \
  return Qnil; \
};

#define FN_S_NEW_TIMER(klass) \
VALUE rb_FL##klass##_s_new(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  VALUE rc,rx,ry,rw,rh,rlabel; \
  switch( rb_scan_args(argc, argv, "06", &rc, &rx, &ry, &rw, &rh, &rlabel) ){ \
  case 0: \
    widget = new RBFL##klass(0, 0, 0, 0, 0); \
    break; \
  case 1: \
    widget = new RBFL##klass(rc, 0, 0, 0, 0); \
    break; \
  case 2: \
    rlabel = rx; \
    widget = new RBFL##klass(rc, 0, 0, 0, 0, strdup(STR2CSTR(rlabel))); \
    break; \
  case 3: \
    rw = rx; rh = ry; \
    widget = new RBFL##klass(rc, 0, 0, NUM2INT(rw), NUM2INT(rh)); \
    break; \
  case 4: \
    rlabel = rw; rw = rx; rh = ry; \
    widget = new RBFL##klass(rc, 0, 0, NUM2INT(rw), NUM2INT(rh), \
                             strdup(STR2CSTR(rlabel))); \
    break; \
  case 5: \
    widget = new RBFL##klass(rc, \
                             NUM2INT(rx), NUM2INT(ry), \
			     NUM2INT(rw), NUM2INT(rh)); \
    break; \
  case 6: \
    widget = new RBFL##klass(rc, \
                             NUM2INT(rx), NUM2INT(ry), \
			     NUM2INT(rw), NUM2INT(rh), \
                             strdup(STR2CSTR(rlabel))); \
    break; \
  }; \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_HANDLE(klass) \
VALUE rb_FL##klass##_##handle(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2NUM(widget->call_parent_handle(NUM2INT(arg1))); \
};


#define FN_RESIZE(klass) \
VALUE rb_FL##klass##_##resize(VALUE self, VALUE x, VALUE y, VALUE w, VALUE h){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  widget->call_parent_resize(NUM2INT(x), NUM2INT(y), \
                     NUM2INT(w), NUM2INT(h)); \
  return Qnil; \
};

#define FN_FLUSH(klass) FN_CALL_RENAME(klass, flush, call_parent_flush)
#define FN_SHOW(klass)  FN_CALL_RENAME(klass, show, call_parent_show)
#define FN_HIDE(klass)  FN_CALL_RENAME(klass, hide, call_parent_hide)
#define FN_DRAW(klass) FN_CALL_RENAME(klass, draw, call_parent_draw)
#define FN_DRAW_OVERLAY(klass) FN_CALL_RENAME(klass, draw_overlay, call_parent_draw_overlay)

#define WIDGET_FNS(klass) \
  static FN_HANDLE(klass) \
  static FN_DRAW(klass) \
  static FN_RESIZE(klass)

#define HELP_FNS(klass) \
  static FN_RESIZE(klass)

#define WINDOW_FNS(klass) \
  WIDGET_FNS(klass) \
  static FN_FLUSH(klass) \
  static FN_SHOW(klass) \
  static FN_HIDE(klass)

#define OVERLAY_WINDOW_FNS(klass) \
  WINDOW_FNS(klass) \
  static FN_DRAW_OVERLAY(klass)

#define FN_MENU_ADD(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  char* str = (arg1 == Qnil) ? NULL : STR2CSTR(arg1); \
  widget->method(str,0,0); \
  return Qnil; \
};

#define FN_BROWSER_TEXT(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass*widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1, arg2; \
  char *str; \
  switch( rb_scan_args(argc, argv, "11", &arg1, &arg2) ){ \
  case 1: \
    str = (char*)(widget->method(NUM2INT(arg1))); \
    return str ? rb_tainted_str_new2(str) : Qnil; \
  case 2: \
    str = (arg2 == Qnil) ? NULL : STR2CSTR(arg2); \
    widget->method(NUM2INT(arg1), str); \
    return Qnil; \
  }; \
};


#define FN_CALL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  widget->method(); \
  return Qnil; \
};

#define FN_CALL_RENAME(klass,rb_method, cpp_method) \
VALUE rb_FL##klass##_##rb_method(VALUE self){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  widget->cpp_method(); \
  return Qnil; \
};

#define FN_GET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return INT2NUM((int)(((RBFL##klass*)widget)->method())); \
};

#define FN_GET_BOOL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  int result = widget->method(); \
  if(result == 0) \
	  return Qfalse; \
  return Qtrue; \
};

#define FN_GET_BOOL_SET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2BOOL(widget->method((int)NUM2INT(arg1))); \
};

#define FN_GET_INT_SET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2NUM(widget->method(NUM2INT(arg1))); \
};

#define FN_GET_INT_SET_INT4(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2, \
                                          VALUE arg3, VALUE arg4){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2NUM(widget->method(NUM2INT(arg1),NUM2INT(arg2), \
                                NUM2INT(arg3),NUM2INT(arg3))); \
};


#define FN_GETSET_ENUM(klass,type,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return INT2NUM(((RBFL##klass*)widget)->method()); \
  case 1: \
    ((RBFL##klass*)widget)->method((type)NUM2INT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_GETSET_ENUM2(klass,type1,type2,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1, arg2; \
  switch( rb_scan_args(argc, argv, "02", &arg1, &arg2) ){ \
  case 0: \
    return INT2NUM(((RBFL##klass*)widget)->method()); \
  case 1: \
    ((RBFL##klass*)widget)->method((type1)NUM2INT(arg1)); \
    return Qnil; \
  case 2: \
    ((RBFL##klass*)widget)->method((type1)NUM2INT(arg1), \
                                   (type2)NUM2INT(arg2)); \
    return Qnil; \
  }; \
};

#define FN_SET_ENUM(klass,type1,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  Fl_Widget *widget = rb_to_fl(self); \
  ((RBFL##klass*)widget)->method((type1)NUM2INT(arg1)); \
  return Qnil; \
};

#define FN_SET_ENUM2(klass,type1,type2,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  Fl_Widget *widget = rb_to_fl(self); \
  ((RBFL##klass*)widget)->method((type1)NUM2INT(arg1), \
			         (type2)NUM2INT(arg2)); \
  return Qnil; \
};

#define FN_SET_ENUM4(klass,type1,type2,type3,type4,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4){ \
  Fl_Widget *widget = rb_to_fl(self); \
  ((RBFL##klass*)widget)->method((type1)NUM2INT(arg1), \
			         (type2)NUM2INT(arg2), \
				 (type3)NUM2INT(arg3), \
				 (type4)NUM2INT(arg4)); \
  return Qnil; \
};

#define FN_SET_INT(klass,method)  FN_SET_ENUM(klass,int,method)
#define FN_SET_INT2(klass,method) FN_SET_ENUM2(klass,int,int,method)
#define FN_SET_INT4(klass,method) FN_SET_ENUM4(klass,int,int,int,int,method)

#define FN_SET_INT_OPTINT(klass, method) \
  VALUE rb_FL##klass##_##method(VALUE argc, VALUE argv[], VALUE self) { \
    RBFL##klass* widget = (RBFL##klass*)rb_to_fl(self); \
    VALUE arg1, arg2; \
    switch( rb_scan_args(argc, argv, "11", &arg1, &arg2) ){ \
      case 1: \
        widget->method(NUM2INT(arg1)); \
        break; \
      case 2: \
        widget->method(NUM2INT(arg1), NUM2INT(arg2)); \
	break; \
    } \
    return Qnil; \
  };

#define FN_SET_INT2_OPTINT3(klass, method) \
  VALUE rb_FL##klass##_##method(VALUE argc, VALUE argv[], VALUE self) { \
    RBFL##klass* widget = (RBFL##klass*)rb_to_fl(self); \
    VALUE arg1, arg2, arg3, arg4, arg5; \
    switch( rb_scan_args(argc, argv, "23", &arg1, &arg2, &arg3, &arg4, &arg5) ){ \
      case 2: \
        widget->method(NUM2INT(arg1), NUM2INT(arg2)); \
		break; \
      case 3: \
        widget->method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3)); \
		break; \
      case 4: \
        widget->method(NUM2INT(arg1), NUM2INT(arg2), \
                       NUM2INT(arg3), NUM2INT(arg4)); \
		break; \
      case 5: \
        widget->method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), \
                       NUM2INT(arg4), NUM2INT(arg5)); \
		break; \
    } \
    return Qnil; \
  };
  
#define FN_GETSET_BOOL(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return INT2BOOL(((RBFL##klass*)widget)->method()); \
  case 1: \
    ((RBFL##klass*)widget)->method(BOOL2INT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_GETSET_INT(klass,method) FN_GETSET_ENUM(klass,int,method)

#define FN_GETSET_DBL(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_float_new(((RBFL##klass*)widget)->method()); \
  case 1: \
    arg1 = rb_Float(arg1); \
    ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value); \
  return Qnil; \
  }; \
};

#define FN_SET_DBL2(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  Fl_Widget *widget = rb_to_fl(self); \
  arg1 = rb_Float(arg1); arg2 = rb_Float(arg2); \
  ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value, RFLOAT(arg2)->value); \
  return Qnil; \
};

#define FN_GET_DBL_SET_DBL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  Fl_Widget *widget = rb_to_fl(self); \
  double ret; \
  arg1 = rb_Float(arg1); \
  ret = ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value); \
  return rb_float_new(ret); \
};

#define FN_GET_DBL_SET_DBL_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  Fl_Widget *widget = rb_to_fl(self); \
  double ret; \
  arg1 = rb_Float(arg1); \
  ret = ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value, NUM2INT(arg2)); \
  return rb_float_new(ret); \
};

#define FN_SET_INT_RENAME(klass,rb_method,cpp_method) \
VALUE rb_FL##klass##_##rb_method(VALUE self, VALUE arg1){ \
  Fl_Widget *widget = rb_to_fl(self); \
  ((RBFL##klass*)widget)->cpp_method(NUM2INT(arg1)); \
  return Qnil; \
};

#define FN_GETSET_INT_RENAME(klass,rb_method,cpp_method) \
VALUE rb_FL##klass##_##rb_method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return INT2NUM(((RBFL##klass*)widget)->cpp_method()); \
  case 1: \
    ((RBFL##klass*)widget)->cpp_method(NUM2INT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_GET_STR(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return rb_tainted_str_new2(((RBFL##klass*)widget)->method()); \
};

#define FN_GET_STR_SET_OPTINT(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_tainted_str_new2(((RBFL##klass*)widget)->method()); \
  case 1: \
    return rb_tainted_str_new2(((RBFL##klass*)widget)->method(NUM2INT(arg1))); \
  }; \
};

#define FN_GET_INT_SET_STR(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE str){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return INT2NUM((int)(((RBFL##klass*)widget)->method(STR2CSTR(str)))); \
};

#define FN_GETSET_STR(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass*widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  char *str; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    str = (char*)(widget->method()); \
    return str ? rb_tainted_str_new2(str) : Qnil; \
  case 1: \
    str = (char*)STR2CPTR(arg1); \
    widget->method(str); \
    return Qnil; \
  }; \
};

#define FN_GETSET_STR_COPY(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass*widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  char *str; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    str = (char*)(widget->method()); \
    return str ? rb_tainted_str_new2(str) : Qnil; \
  case 1: \
    str = (char*)(widget->method()); \
    if( str ) free(str); \
    str = (arg1 == Qnil) ? NULL : strdup((char*)STR2CSTR(arg1)); \
    widget->method(str); \
    return Qnil; \
  }; \
};

#define FN_SET_STR_COPY(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass*widget = (RBFL##klass*)rb_to_fl(self); \
  char* str = (arg1 == Qnil) ? NULL : STR2CSTR(arg1); \
  widget->method(str); \
  return Qnil; \
};

#define FN_GET_VALUE(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return (VALUE)(((RBFL##klass*)widget)->method()); \
};

#define FN_GETSET_VALUE(klass,type,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE obj; \
  switch( rb_scan_args(argc, argv, "01", &obj) ){ \
  case 0: \
    return (VALUE)(((RBFL##klass*)widget)->method()); \
  case 1: \
    ((RBFL##klass*)widget)->method((type)obj); \
    return Qnil; \
  }; \
};

#define FN_GET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return rb_fltk_objmap_aref((Fl_Widget*)((RBFL##klass*)widget)->method()); \
};

#define FN_S_GET_WIDGET(klass,method) \
VALUE rb_FL##klass##_s_##method(VALUE self){ \
  return rb_fltk_objmap_aref((Fl_Widget*)(RBFL##klass::method())); \
};

#define FN_GETSET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_fltk_objmap_aref(widget->method()); \
  case 1: \
    widget->method(rb_to_fl(arg1)); \
    return Qnil; \
  }; \
};

#define FN_GET_BOOL_SET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  if( widget->method(rb_to_fl(w)) ){ \
    return Qtrue; \
  } \
  else{ \
    return Qfalse; \
  }; \
};

#define FN_GET_INT_SET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  return INT2NUM(widget->method(rb_to_fl(w))); \
};

#define FN_SET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method(rb_to_fl(w)); \
  return Qnil; \
};

#define FN_SET_REFWIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method(*(rb_to_fl(w))); \
  return Qnil; \
};



#define FN_MODULE_FL_CALL(method) \
VALUE rb_fltk_##method(VALUE self){ \
  Fl::method(); \
  return Qnil; \
};

#define FN_MODULE_FL_GET_INT(method) \
VALUE rb_fltk_##method(VALUE self){ \
  return INT2NUM(Fl::method()); \
};

#define FN_MODULE_FL_GET_INT_SET_INT(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1){ \
  return INT2NUM(Fl::method(NUM2INT(arg1))); \
};

#define FN_MODULE_FL_GETSET_ENUM(etype,method) \
VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self){ \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return INT2NUM((int)(Fl::method())); \
  case 1: \
    Fl::method((etype)NUM2INT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_MODULE_FL_GETSET_INT(method) FN_MODULE_FL_GETSET_ENUM(int,method)

#define FN_MODULE_FL_GET_STR(method) \
VALUE rb_fltk_##method(VALUE self){ \
  return rb_tainted_str_new2(Fl::method()); \
};

#define FN_MODULE_FL_GETSET_STR(method) \
VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self){ \
  VALUE arg1; \
  char *str; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_tainted_str_new2(Fl::method()); \
  case 1: \
    str = Fl::method(); \
    if( str ) free(str); \
    Fl::method(strdup(STR2CSTR(arg1))); \
    return Qnil; \
  }; \
};

#define FN_GLOBAL_CALL(method) \
VALUE rb_fltk_##method(VALUE self){ \
  fl_##method(); \
  return Qnil; \
};

#define FN_GL_GLOBAL_CALL(method) \
VALUE rb_fltk_##gl_##method(VALUE self){ \
  gl_##method(); \
  return Qnil; \
};

#define FN_GLOBAL_GET_BOOL_SET_INT4(method) \
VALUE rb_fltk_##method(VALUE mod, VALUE x, VALUE y, VALUE w, VALUE h){ \
  return fl_##method(NUM2INT(x),NUM2INT(y),NUM2INT(w),NUM2INT(h)) ? Qtrue : Qfalse; \
};


#define FN_GLOBAL_GETSET_INT(method) \
  VALUE rb_fltk_##method(VALUE argc, VALUE argv[], VALUE self) { \
    VALUE arg1; \
    switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
      case 0: \
        return INT2NUM(fl_##method()); \
      case 1: \
        fl_##method(NUM2INT(arg1)); \
		break; \
    } \
    return Qnil; \
  };
  
#define FN_GLOBAL_SET_ENUM(method,type) \
  VALUE rb_fltk_##method(VALUE self, VALUE x) { \
    fl_##method((type)NUM2INT(x)); \
    return Qnil; \
  };

#define FN_GL_GLOBAL_SET_ENUM(method,type) \
  VALUE rb_fltk_##gl_##method(VALUE self, VALUE x) { \
    gl_##method((type)NUM2INT(x)); \
    return Qnil; \
  };

#define FN_GLOBAL_SET_INT(method) FN_GLOBALE_SET_ENUM(method,int)
#define FN_GL_GLOBAL_SET_INT(method) FN_GL_GLOBAL_SET_ENUM(method,int)

#define FN_GLOBAL_SET_INT2(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE x, VALUE y) { \
    fl_##method(NUM2INT(x), NUM2INT(y)); \
    return Qnil; \
  };

#define FN_GL_GLOBAL_SET_INT2(method) \
  VALUE rb_fltk_##gl_##method(VALUE self, VALUE x, VALUE y) { \
    gl_##method(NUM2INT(x), NUM2INT(y)); \
    return Qnil; \
  };
  
#define FN_GLOBAL_SET_INT4_OPTINT2(method) \
  VALUE rb_fltk_##method(VALUE argc, VALUE argv[], VALUE self) { \
    VALUE arg1, arg2, arg3, arg4, arg5, arg6; \
    switch( rb_scan_args(argc, argv, "42", &arg1, &arg2, &arg3, &arg4, &arg5, &arg6) ){ \
      case 4: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4)); \
		break; \
      case 6: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4), \
		           NUM2INT(arg5), NUM2INT(arg6)); \
		break; \
      default: \
	 rb_raise(rb_eArgError, "too few/many arguments"); \
         break; \
    } \
    return Qnil; \
  };
  
#define FN_GLOBAL_SET_INT4(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE x, VALUE y, VALUE w, VALUE h) { \
    fl_##method(NUM2INT(x), NUM2INT(y), NUM2INT(w), NUM2INT(h)); \
    return Qnil; \
  };
  
#define FN_GLOBAL_SET_INT4_OPTCHR3(method) \
  VALUE rb_fltk_##method(VALUE argc, VALUE argv[], VALUE self) { \
    VALUE arg1, arg2, arg3, arg4, arg5, arg6, arg7; \
    switch( rb_scan_args(argc, argv, "42", &arg1, &arg2, &arg3, &arg4, \
	             &arg5, &arg6, arg7) ){ \
      case 4: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4)); \
		break; \
      case 7: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4), \
		           NUM2CHR(arg5), NUM2CHR(arg6), NUM2CHR(arg7)); \
		break; \
      default: \
	    /* TODO: should raise an error here */ \
	    break; \
    } \
    return Qnil; \
  };
  
#define FN_GLOBAL_SET_INT4_DBL2(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE x, VALUE y, VALUE w, VALUE h, VALUE a1, VALUE a2) { \
    fl_##method(NUM2INT(x), NUM2INT(y), NUM2INT(w), NUM2INT(h), \
		           NUM2DBL(a1), NUM2DBL(a2)); \
    return Qnil; \
  };

#define FN_GLOBAL_SET_DBL2(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  fl_##method(NUM2DBL(arg1), NUM2DBL(arg2)); \
  return Qnil; \
}

#define FN_GLOBAL_SET_DBL3(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3){ \
  fl_##method(NUM2DBL(arg1), NUM2DBL(arg2), NUM2DBL(arg3)); \
  return Qnil; \
}

#define FN_GLOBAL_SET_DBL4(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4){ \
  fl_##method(NUM2DBL(arg1), NUM2DBL(arg2), NUM2DBL(arg3), NUM2DBL(arg4)); \
  return Qnil; \
}
  
#define FN_GLOBAL_SET_STR(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE arg1) { \
    char* str = (arg1 == Qnil) ? NULL : STR2CSTR(arg1); \
    fl_##method("%s", str); \
    return Qnil; \
  };

#define FN_GLOBAL_SET_FORMAT(method) \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE val; \
    char *str; \
    val = rb_f_sprintf(argc, argv); \
    str = (val == Qnil) ? NULL : STR2CSTR(val); \
    fl_##method("%s", str); \
    return Qnil; \
  };

#define FN_GLOBAL_GET_BOOL_SET_FORMAT(method) \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE val; \
    char *str; \
    int ret; \
    val = rb_f_sprintf(argc, argv); \
    str = (val == Qnil) ? NULL : STR2CSTR(val); \
    if( fl_##method("%s", str) ){ \
      return Qtrue; \
    } \
    else{ \
      return Qfalse; \
    }; \
  };

#define FN_GLOBAL_GET_STR_SET_FORMAT_STR(method) \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE arg1, arg2; \
    char *fmt, *deflt; \
    const char *result; \
    int ret; \
    switch( rb_scan_args(argc, argv, "11", &arg1, &arg2) ){ \
    case 1: \
      fmt = STR2CSTR(arg1); \
      deflt = NULL; \
      break; \
    case 2: \
      fmt = STR2CSTR(arg1); \
      deflt = STR2CPTR(arg2); \
      break; \
    }; \
    if( result = fl_##method("%s", deflt, fmt) ){ \
      return rb_tainted_str_new2(result); \
    } \
    else{ \
      return Qnil; \
    }; \
  };

#define FN_GLOBAL_GET_INT_SET_FORMAT_STR3(method) \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE fmt  = Qnil; \
    VALUE str1 = Qnil, str2 = Qnil, str3 = Qnil; \
    int ret; \
    switch( rb_scan_args(argc, argv, "13", &fmt, &str1, &str2, &str3) ){ \
    case 1: \
      ret = fl_##method("%s",NULL,NULL,NULL,STR2CSTR(fmt)); \
      break; \
    case 2: \
      ret = fl_##method("%s",STR2CPTR(str1),NULL,NULL,STR2CSTR(fmt)); \
      break; \
    case 3: \
      ret = fl_##method("%s",STR2CPTR(str1),STR2CPTR(str2),NULL,STR2CSTR(fmt)); \
      break; \
    case 4: \
      ret = fl_##method("%s",STR2CPTR(str1), STR2CPTR(str2), \
                             STR2CPTR(str3), STR2CPTR(fmt)); \
      break; \
    default: \
      rb_raise(rb_eArgError, "too many arguments"); \
    }; \
    return INT2NUM(ret); \
  };


/* wrappers for rb_define_{method,singleton_method,module_function} */
#define RB_DEF_METHOD(klass,method,func,argtype) \
        rb_define_method(klass,method,RUBY_METHOD_FUNC(func), argtype)
#define RB_DEF_S_METHOD(klass,method,func,argtype) \
        rb_define_singleton_method(klass,method,RUBY_METHOD_FUNC(func), argtype)
#define RB_DEF_MODULE_FUNC(module,method,func,argtype) \
        rb_define_module_function(module,method,RUBY_METHOD_FUNC(func), argtype)

#define DEF_S_METHOD(klass,method,argtype) \
        RB_DEF_S_METHOD(rb_cFL##klass, #method, rb_FL##klass##_s_##method, argtype)
#define DEF_METHOD(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype)
#define DEF_METHOD_Q(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #method"?", rb_FL##klass##_##method, argtype)
#define DEF_METHOD_EQ(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #method"=", rb_FL##klass##_##method, argtype)
#define DEF_METHOD_QEQ(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #method"=", rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #method"?", rb_FL##klass##_##method, argtype)

#define DEF_PREFIXED_METHOD(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype)
#define DEF_PREFIXED_METHOD_Q(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"?", \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"?", \
		      rb_FL##klass##_##method, argtype)
#define DEF_PREFIXED_METHOD_EQ(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"=", \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"=", \
		      rb_FL##klass##_##method, argtype)
#define DEF_PREFIXED_METHOD_QEQ(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"=", \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"=", \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"?", \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"?", \
		      rb_FL##klass##_##method, argtype)

#define DEF_MODULE_FUNC(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype)
#define DEF_MODULE_FUNC_Q(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"?", rb_fltk_##method, argtype)
#define DEF_MODULE_FUNC_EQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"=", rb_fltk_##method, argtype)
#define DEF_MODULE_FUNC_QEQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"?", rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"=", rb_fltk_##method, argtype)

#define DEF_GL_MODULE_FUNC(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype)
#define DEF_GL_MODULE_FUNC_Q(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"?", rb_fltk_##gl_##method, argtype)
#define DEF_GL_MODULE_FUNC_EQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"=", rb_fltk_##gl_##method, argtype)
#define DEF_GL_MODULE_FUNC_QEQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"?", rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"=", rb_fltk_##gl_##method, argtype)

#define DEF_WIDGET_METHODS(klass) \
  DEF_METHOD(klass, handle, 1); \
  DEF_METHOD(klass, draw, 0); \
  DEF_METHOD(klass, resize, 4);

#define DEF_HELP_METHODS(klass) \
  DEF_METHOD(klass, resize, 4);

#define DEF_WINDOW_METHODS(klass) \
  DEF_WIDGET_METHODS(klass); \
  DEF_METHOD(klass, flush, 0); \
  DEF_METHOD(klass, show, 0); \
  DEF_METHOD(klass, hide, 0);

#define DEF_OVERLAY_WINDOW_METHODS(klass) \
  DEF_WINDOW_METHODS(klass); \
  DEF_METHOD(klass, draw_overlay, 0); \

#endif /* RBFLTK_H */
