/* -*- C++ -*-
 * $Id: flwidget.cc,v 1.59 2001/11/30 06:08:04 ttate Exp $
 */

#include "fltk.h"
#include "flwidget.h"
#include "flgroup.h"

VALUE rb_cFLWidget;

static void
rb_fl_free(Fl_Widget *widget)
{
  /* I assume that 'delete' is done by Widget#destroy method. */
};

static void
rb_fl_mark(Fl_Widget *widget)
{
  VALUE obj;
  void *userdata;

  if( !widget ){
    return;
  };
  if( rb_fltk_objmap_aref(widget) == Qnil ){
    return;
  };

  if( userdata = widget->user_data() ){
    rb_gc_mark((VALUE)userdata);
  };

  obj = rb_fltk_objmap_aref(widget);
  if( rb_obj_is_kind_of(obj, rb_cFLGroup) ){
    Fl_Widget *child;
    VALUE child_obj;
    int n = ((RBFLGroup*)widget)->children();
    for( int i = 0; i < n; i++ ){
      child = ((RBFLGroup*)widget)->child(i);
      child_obj = rb_fltk_objmap_aref(child);
      if( child_obj != Qnil ){
	rb_fl_mark(rb_to_fl(child_obj));
      };
    };
  };
};

VALUE
rb_fl_new(VALUE klass, Fl_Widget *widget)
{
  VALUE obj;
  
  obj = Data_Wrap_Struct(klass, rb_fl_mark, rb_fl_free, widget);
  rb_fltk_objmap_aset(widget, obj);
  rb_ivar_set(obj, rb_id_icallback, Qnil);

  return obj;
};

static void
rb_fl_destroy_internal(VALUE obj)
{
  Fl_Widget *widget;

  if( obj == Qnil ) return ;

  widget = rb_to_fl(obj);
  rb_fltk_objmap_delete(widget);
  if( rb_obj_is_kind_of(obj, rb_cFLGroup) ){
    Fl_Widget *child;
    int n = ((RBFLGroup*)widget)->children();
    for( int i = 0; i < n; i++ ){
      child = ((RBFLGroup*)widget)->child(i);
      rb_fl_destroy_internal(rb_fltk_objmap_aref(child));
    };
  };
};

VALUE
rb_fl_destroy(VALUE self)
{
  Fl_Widget *widget;

  widget = rb_to_fl(self);
  rb_fl_destroy_internal(self);
  if( widget )
    delete widget;
  
  return Qnil;
};

VALUE
rb_fl_destroyed(VALUE self)
{
  Fl_Widget *widget;
  
  widget = (Fl_Widget*)DATA_PTR(self);
  if( rb_fltk_objmap_aref(widget) == Qnil ){
    return Qtrue;
  }
  else{
    return Qfalse;
  };
};  

VALUE
rb_fl_valid(VALUE self)
{
  Fl_Widget *widget;

  rb_warn("use 'destroyed?' instead of 'valid?'.");
  widget = (Fl_Widget*)DATA_PTR(self);
  if( rb_fltk_objmap_aref(widget) == Qnil ){
    return Qtrue;
  }
  else{
    return Qfalse;
  };
};  

VALUE
rb_fl_check(VALUE self)
{
  if( rb_fl_destroyed(self) == Qtrue ){
    rb_raise(rb_eFLError, "the widget have been destroied.");
  };
  return Qnil;
};

VALUE
rb_fl_check_widget(VALUE self)
{
  if( !rb_obj_is_kind_of(self, rb_cFLWidget) ){
    rb_raise(rb_eTypeError, "%s is not a FLWidget.",
	     STR2CSTR(rb_funcall(self, rb_intern("inspect"), 0)));
  };
  return Qnil;
};

Fl_Widget *
rb_to_fl(VALUE self)
{
  if( self == Qnil ){
    return NULL;
  };
  rb_fl_check(self);
  return (Fl_Widget*)DATA_PTR(self);
};

static void
rb_flwidget_callback_func(Fl_Widget *widget, void *userdata)
{
  VALUE obj;
  VALUE callback, data;

  obj = rb_fltk_objmap_aref(widget);
  if( obj == Qnil ) return ;

  callback = rb_ivar_get(obj, rb_id_icallback);
  if( userdata ){
    data = (VALUE)userdata;
  }
  else{
    data = Qnil;
  };

  if( callback != Qnil ){
    rb_funcall(callback, rb_id_call, 2, obj, data);
  };
};

static VALUE
rb_FLWidget_callback(int argc, VALUE argv[], VALUE self)
{
  VALUE proc, data;
  Fl_Widget *widget;
  int use_data = 0;

  if( rb_block_given_p() ){
    proc = rb_f_lambda();
    switch( rb_scan_args(argc, argv, "01", &data) ){
    case 0:
      use_data = 0;
      break;
    case 1:
      use_data = 1;
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "11", &proc, &data) ){
    case 1:
      use_data = 0;
      break;
    case 2:
      use_data = 1;
      break;
    };
  };

  rb_ivar_set(self, rb_id_icallback, proc);
  widget = rb_to_fl(self);
  if( use_data ){
    ((RBFLWidget*)widget)->callback((Fl_Callback*)rb_flwidget_callback_func,
				    (void*)data);
  }
  else{
    ((RBFLWidget*)widget)->callback((Fl_Callback*)rb_flwidget_callback_func);
  };

  return Qnil;
};

VALUE
rb_fl_callback(VALUE self)
{
  VALUE argv[1] = {Qnil};

  return rb_FLWidget_callback(0, argv, self);
};

VALUE
rb_FLWidget_user_data(int argc, VALUE argv[], VALUE self)
{
  Fl_Widget *widget = rb_to_fl(self);
  VALUE data;

  switch( rb_scan_args(argc, argv, "01", &data) ){
  case 0:
    return (VALUE)(widget->user_data());
  case 1:
    widget->user_data((void*)data);
    return Qnil;
  };
};

static VALUE rb_FLWidget_do_callback(int argc, VALUE argv[], VALUE self)
{
  Fl_Widget *widget = rb_to_fl(self);
  VALUE w, data;

  switch( rb_scan_args(argc, argv, "02", &w, &data) ){
  case 2:
    widget->do_callback(rb_to_fl(w), (void*)data);
    break;
  case 1:
    widget->do_callback(rb_to_fl(w));
    break;
  case 0:
    widget->do_callback();
    break;
  };
  return Qnil;
};

static FN_S_NEW(Widget);
static FN_INIT(Widget);
WIDGET_FNS(Widget);
static FN_GETSET_INT(Widget,type);
static FN_GET_INT(Widget,x);
static FN_GET_INT(Widget,y);
static FN_GET_INT(Widget,w);
static FN_GET_INT(Widget,h);
static FN_SET_INT2(Widget,position);
static FN_SET_INT2(Widget,size);
static FN_GET_WIDGET(Widget,window);
static FN_GET_WIDGET(Widget,parent);
static FN_GETSET_ENUM(Widget,Fl_Boxtype,box);
static FN_GETSET_ENUM2(Widget,Fl_Color,Fl_Color,color);
static FN_GETSET_ENUM(Widget,Fl_Color,selection_color);
static FN_GETSET_STR_COPY(Widget,label);
static FN_GETSET_ENUM(Widget,Fl_Labeltype,labeltype);
static FN_GETSET_ENUM(Widget,Fl_Font,labelfont);
static FN_GETSET_ENUM(Widget,Fl_Color,labelcolor);
static FN_GETSET_INT(Widget,labelsize);
static FN_GETSET_ENUM(Widget,Fl_Align,align);
static FN_CALL(Widget,show);
static FN_CALL(Widget,hide);
static FN_CALL(Widget,redraw);
static FN_GETSET_ENUM(Widget,Fl_When,when);
static FN_GET_BOOL(Widget,visible);
static FN_GET_BOOL(Widget,visible_r);
static FN_GET_BOOL(Widget,active);
static FN_GET_BOOL(Widget,active_r);
static FN_CALL(Widget,activate);
static FN_CALL(Widget,deactivate);
static FN_GET_BOOL(Widget,output);
static FN_CALL(Widget,set_output);
static FN_CALL(Widget,clear_output);
static FN_GET_BOOL(Widget,takesevents);
static FN_GET_BOOL(Widget,damage);
static FN_GET_BOOL_SET_WIDGET(Widget,contains);
static FN_GET_BOOL_SET_WIDGET(Widget,inside);
static FN_GET_BOOL(Widget,take_focus);

extern "C" void
Init_flwidget(){
  rb_cFLWidget = rb_define_class_under(rb_mFLTK, "Widget", rb_cObject);

  DEF_S_METHOD(Widget, new, -1);
  DEF_METHOD(Widget,initialize,-1);

  RB_DEF_METHOD(rb_cFLWidget, "destroy", rb_fl_destroy, 0);
  RB_DEF_METHOD(rb_cFLWidget, "destroyed?", rb_fl_destroyed, 0);
  RB_DEF_METHOD(rb_cFLWidget, "valid?", rb_fl_valid, 0);
  DEF_METHOD(Widget,callback,-1);
  DEF_METHOD_EQ(Widget,user_data,-1);

  DEF_WIDGET_METHODS(Widget);
  DEF_PREFIXED_METHOD_EQ(Widget,widget,type,-1);
  DEF_METHOD(Widget,do_callback,-1);
  DEF_METHOD(Widget,window,0);
  DEF_METHOD(Widget,parent,0);
  DEF_METHOD(Widget,x,0);
  DEF_METHOD(Widget,y,0);
  DEF_METHOD(Widget,w,0);
  DEF_METHOD(Widget,h,0);
  DEF_METHOD(Widget,position,2);
  DEF_METHOD(Widget,size,2);
  DEF_METHOD_EQ(Widget,box,-1);
  DEF_METHOD_EQ(Widget,color,-1);
  DEF_METHOD_EQ(Widget,selection_color,-1);
  DEF_METHOD_EQ(Widget,label,-1);
  DEF_METHOD_EQ(Widget,labeltype,-1);
  DEF_METHOD_EQ(Widget,labelfont,-1);
  DEF_METHOD_EQ(Widget,labelcolor,-1);
  DEF_METHOD_EQ(Widget,labelsize,-1);
  DEF_METHOD(Widget,show,0);
  DEF_METHOD(Widget,hide,0);
  DEF_METHOD(Widget,redraw,0);
  DEF_METHOD(Widget,when,-1);
  DEF_METHOD_EQ(Widget,align,-1);
  DEF_METHOD_Q(Widget,visible,0);
  DEF_METHOD_Q(Widget,visible_r,0);
  DEF_METHOD_Q(Widget,active,0);
  DEF_METHOD_Q(Widget,active_r,0);
  DEF_METHOD(Widget,activate,0);
  DEF_METHOD(Widget,deactivate,0);
  DEF_METHOD_Q(Widget,output,0);
  DEF_METHOD(Widget,set_output,0);
  DEF_METHOD(Widget,clear_output,0);
  DEF_METHOD_Q(Widget,takesevents,0);
  DEF_METHOD_Q(Widget,damage,0);
  DEF_METHOD_Q(Widget,contains,1);
  DEF_METHOD_Q(Widget,inside,1);
  DEF_METHOD_Q(Widget,take_focus,0);
};
